# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from typing import Dict, Optional

import requests

from lightning.app.cli.connect.app import (
    _clean_lightning_connection,
    _install_missing_requirements,
    _resolve_command_path,
)
from lightning.app.utilities.cli_helpers import _LightningAppOpenAPIRetriever
from lightning.app.utilities.commands.base import _download_command
from lightning.app.utilities.enum import OpenAPITags


def _is_running_help(argv) -> bool:
    return argv[-1] in ["--help", "-"] if argv else False


def _run_app_command(app_name: str, app_id: Optional[str]):
    """Execute a function in a running App from its name."""
    # 1: Collect the url and comments from the running application
    _clean_lightning_connection()

    running_help = _is_running_help(sys.argv)

    retriever = _LightningAppOpenAPIRetriever(app_id, use_cache=running_help)

    if not running_help and (retriever.url is None or retriever.api_commands is None):
        if app_name == "localhost":
            print("The command couldn't be executed as your local Lightning App isn't running.")
        else:
            print(f"The command couldn't be executed as your cloud Lightning App `{app_name}` isn't running.")
        sys.exit(0)

    if not retriever.api_commands:
        raise Exception("This application doesn't expose any commands yet.")

    full_command = "_".join(sys.argv)

    has_found = False
    for command in list(retriever.api_commands):
        if command in full_command:
            has_found = True
            for value in sys.argv:
                if value == command and "_" in value:
                    print(
                        f"The command `{value}` was provided with an underscore and it isn't allowed."
                        f"Instead, use `lightning {value.replace('_', ' ')}`."
                    )
                    sys.exit(0)
            break

    if not has_found:
        raise Exception(f"The provided command isn't available in {list(retriever.api_commands)}")

    # 2: Send the command from the user
    metadata = retriever.api_commands[command]

    try:
        # 3: Execute the command
        if metadata["tag"] == OpenAPITags.APP_COMMAND:
            _handle_command_without_client(command, metadata, retriever.url)
        else:
            _handle_command_with_client(command, metadata, app_name, app_id, retriever.url)
    except ModuleNotFoundError:
        _install_missing_requirements(retriever, fail_if_missing=True)

    if running_help:
        print("Your command execution was successful.")


def _handle_command_without_client(command: str, metadata: Dict, url: str) -> None:
    supported_params = list(metadata["parameters"])
    if _is_running_help(sys.argv):
        print(f"Usage: lightning {command} [ARGS]...")
        print(" ")
        print("Options")
        for param in supported_params:
            print(f"  {param}: Add description")
        return

    provided_params = [param.replace("--", "") for param in sys.argv[1 + len(command.split("_")) :]]

    # TODO: Add support for more argument types.
    if any("=" not in param for param in provided_params):
        raise Exception("Please, use --x=y syntax when providing the command arguments.")

    if any(param.split("=")[0] not in supported_params for param in provided_params):
        raise Exception(f"Some arguments need to be provided. The keys are {supported_params}.")

    # TODO: Encode the parameters and validate their type.
    query_parameters = "&".join(provided_params)
    resp = requests.post(url + f"/command/{command}?{query_parameters}")
    assert resp.status_code == 200, resp.json()
    print(resp.json())


def _handle_command_with_client(command: str, metadata: Dict, app_name: str, app_id: Optional[str], url: str):
    debug_mode = bool(int(os.getenv("DEBUG", "0")))

    if app_name == "localhost":
        target_file = metadata["cls_path"]
    else:
        target_file = _resolve_command_path(command) if debug_mode else _resolve_command_path(command)

    if debug_mode:
        print(target_file)

    client_command = _download_command(
        command,
        metadata["cls_path"],
        metadata["cls_name"],
        app_id,
        debug_mode=debug_mode,
        target_file=target_file if debug_mode else _resolve_command_path(command),
    )
    client_command._setup(command_name=command, app_url=url)
    sys.argv = sys.argv[len(command.split("_")) :]
    client_command.run()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from lightning_cloud.openapi import V1Membership

import lightning.app
from lightning.app.core.constants import LIGHTNING_CLOUD_PROJECT_ID
from lightning.app.utilities.enum import AppStage
from lightning.app.utilities.network import LightningClient


def _get_project(
    client: LightningClient, project_id: str = LIGHTNING_CLOUD_PROJECT_ID, verbose: bool = True
) -> V1Membership:
    """Get a project membership for the user from the backend."""
    projects = client.projects_service_list_memberships()
    if project_id is not None:
        for membership in projects.memberships:
            if membership.project_id == project_id:
                break
        else:
            raise ValueError(
                "Environment variable `LIGHTNING_CLOUD_PROJECT_ID` is set but could not find an associated project."
            )
        return membership

    if len(projects.memberships) == 0:
        raise ValueError("No valid projects found. Please reach out to lightning.ai team to create a project")
    if len(projects.memberships) > 1:
        if verbose:
            print(f"Defaulting to the project: {projects.memberships[0].name}")
    return projects.memberships[0]


def _sigterm_flow_handler(*_, app: "lightning.app.LightningApp"):
    app.stage = AppStage.STOPPING


def is_running_in_cloud() -> bool:
    """Returns True if the Lightning App is running in the cloud."""
    return bool(int(os.environ.get("LAI_RUNNING_IN_CLOUD", "0"))) or "LIGHTNING_APP_STATE_URL" in os.environ



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model Summary
=============

Generates a summary of all layers in a :class:`~lightning.pytorch.core.module.LightningModule`.

The string representation of this summary prints a table with columns containing
the name, type and number of parameters for each layer.

"""
import logging
from typing import List, Tuple, Union

import lightning.pytorch as pl
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.utilities.model_summary import DeepSpeedSummary
from lightning.pytorch.utilities.model_summary import ModelSummary as Summary
from lightning.pytorch.utilities.model_summary import summarize
from lightning.pytorch.utilities.model_summary.model_summary import _format_summary_table

log = logging.getLogger(__name__)


class ModelSummary(Callback):
    r"""
    Generates a summary of all layers in a :class:`~lightning.pytorch.core.module.LightningModule`.

    Args:
        max_depth: The maximum depth of layer nesting that the summary will include. A value of 0 turns the
            layer summary off.

    Example::

        >>> from lightning.pytorch import Trainer
        >>> from lightning.pytorch.callbacks import ModelSummary
        >>> trainer = Trainer(callbacks=[ModelSummary(max_depth=1)])
    """

    def __init__(self, max_depth: int = 1) -> None:
        self._max_depth: int = max_depth

    def on_fit_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if not self._max_depth:
            return None

        model_summary = self._summary(trainer, pl_module)
        summary_data = model_summary._get_summary_data()
        total_parameters = model_summary.total_parameters
        trainable_parameters = model_summary.trainable_parameters
        model_size = model_summary.model_size

        if trainer.is_global_zero:
            self.summarize(summary_data, total_parameters, trainable_parameters, model_size)

    def _summary(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> Union[DeepSpeedSummary, Summary]:
        from lightning.pytorch.strategies.deepspeed import DeepSpeedStrategy

        if isinstance(trainer.strategy, DeepSpeedStrategy) and trainer.strategy.zero_stage_3:
            return DeepSpeedSummary(pl_module, max_depth=self._max_depth)
        return summarize(pl_module, max_depth=self._max_depth)

    @staticmethod
    def summarize(
        summary_data: List[Tuple[str, List[str]]],
        total_parameters: int,
        trainable_parameters: int,
        model_size: float,
    ) -> None:
        summary_table = _format_summary_table(total_parameters, trainable_parameters, model_size, *summary_data)
        log.info("\n" + summary_table)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Timer
^^^^^
"""
import logging
import time
from datetime import timedelta
from typing import Any, Dict, Optional, Union

import lightning.pytorch as pl
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.trainer.states import RunningStage
from lightning.pytorch.utilities import LightningEnum
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.rank_zero import rank_zero_info

log = logging.getLogger(__name__)


class Interval(LightningEnum):
    step = "step"
    epoch = "epoch"


class Timer(Callback):
    """The Timer callback tracks the time spent in the training, validation, and test loops and interrupts the
    Trainer if the given time limit for the training loop is reached.

    Args:
        duration: A string in the format DD:HH:MM:SS (days, hours, minutes seconds), or a :class:`datetime.timedelta`,
            or a dict containing key-value compatible with :class:`~datetime.timedelta`.
        interval: Determines if the interruption happens on epoch level or mid-epoch.
            Can be either ``"epoch"`` or ``"step"``.
        verbose: Set this to ``False`` to suppress logging messages.

    Raises:
        MisconfigurationException:
            If ``interval`` is not one of the supported choices.

    Example::

        from lightning.pytorch import Trainer
        from lightning.pytorch.callbacks import Timer

        # stop training after 12 hours
        timer = Timer(duration="00:12:00:00")

        # or provide a datetime.timedelta
        from datetime import timedelta
        timer = Timer(duration=timedelta(weeks=1))

        # or provide a dictionary
        timer = Timer(duration=dict(weeks=4, days=2))

        # force training to stop after given time limit
        trainer = Trainer(callbacks=[timer])

        # query training/validation/test time (in seconds)
        timer.time_elapsed("train")
        timer.start_time("validate")
        timer.end_time("test")
    """

    def __init__(
        self,
        duration: Optional[Union[str, timedelta, Dict[str, int]]] = None,
        interval: str = Interval.step,
        verbose: bool = True,
    ) -> None:
        super().__init__()
        if isinstance(duration, str):
            dhms = duration.strip().split(":")
            dhms = [int(i) for i in dhms]
            duration = timedelta(days=dhms[0], hours=dhms[1], minutes=dhms[2], seconds=dhms[3])
        if isinstance(duration, dict):
            duration = timedelta(**duration)
        if interval not in set(Interval):
            raise MisconfigurationException(
                f"Unsupported parameter value `Timer(interval={interval})`. Possible choices are:"
                f" {', '.join(set(Interval))}"
            )
        self._duration = duration.total_seconds() if duration is not None else None
        self._interval = interval
        self._verbose = verbose
        self._start_time: Dict[RunningStage, Optional[float]] = {stage: None for stage in RunningStage._without_tune()}
        self._end_time: Dict[RunningStage, Optional[float]] = {stage: None for stage in RunningStage._without_tune()}
        self._offset = 0

    def start_time(self, stage: str = RunningStage.TRAINING) -> Optional[float]:
        """Return the start time of a particular stage (in seconds)"""
        stage = RunningStage(stage)
        return self._start_time[stage]

    def end_time(self, stage: str = RunningStage.TRAINING) -> Optional[float]:
        """Return the end time of a particular stage (in seconds)"""
        stage = RunningStage(stage)
        return self._end_time[stage]

    def time_elapsed(self, stage: str = RunningStage.TRAINING) -> float:
        """Return the time elapsed for a particular stage (in seconds)"""
        start = self.start_time(stage)
        end = self.end_time(stage)
        offset = self._offset if stage == RunningStage.TRAINING else 0
        if start is None:
            return offset
        if end is None:
            return time.monotonic() - start + offset
        return end - start + offset

    def time_remaining(self, stage: str = RunningStage.TRAINING) -> Optional[float]:
        """Return the time remaining for a particular stage (in seconds)"""
        if self._duration is not None:
            return self._duration - self.time_elapsed(stage)

    def on_train_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._start_time[RunningStage.TRAINING] = time.monotonic()

    def on_train_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._end_time[RunningStage.TRAINING] = time.monotonic()

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._start_time[RunningStage.VALIDATING] = time.monotonic()

    def on_validation_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._end_time[RunningStage.VALIDATING] = time.monotonic()

    def on_test_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._start_time[RunningStage.TESTING] = time.monotonic()

    def on_test_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._end_time[RunningStage.TESTING] = time.monotonic()

    def on_fit_start(self, trainer: "pl.Trainer", *args: Any, **kwargs: Any) -> None:
        # this checks the time after the state is reloaded, regardless of the interval.
        # this is necessary in case we load a state whose timer is already depleted
        if self._duration is None:
            return
        self._check_time_remaining(trainer)

    def on_train_batch_end(self, trainer: "pl.Trainer", *args: Any, **kwargs: Any) -> None:
        if self._interval != Interval.step or self._duration is None:
            return
        self._check_time_remaining(trainer)

    def on_train_epoch_end(self, trainer: "pl.Trainer", *args: Any, **kwargs: Any) -> None:
        if self._interval != Interval.epoch or self._duration is None:
            return
        self._check_time_remaining(trainer)

    def state_dict(self) -> Dict[str, Any]:
        return {"time_elapsed": {stage.value: self.time_elapsed(stage) for stage in RunningStage._without_tune()}}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        time_elapsed = state_dict.get("time_elapsed", {})
        self._offset = time_elapsed.get(RunningStage.TRAINING.value, 0)

    def _check_time_remaining(self, trainer: "pl.Trainer") -> None:
        assert self._duration is not None
        should_stop = self.time_elapsed() >= self._duration
        should_stop = trainer.strategy.broadcast(should_stop)
        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop and self._verbose:
            elapsed = timedelta(seconds=int(self.time_elapsed(RunningStage.TRAINING)))
            rank_zero_info(f"Time limit reached. Elapsed time is {elapsed}. Signaling Trainer to stop.")



from typing import Dict, Optional, Union

from pydantic import BaseModel, Field


class FitProgress(BaseModel):
    current_epoch: int = 0
    train_batch_idx: int = 0
    total_train_batches: int = 0
    val_dataloader_idx: int = 0
    val_batch_idx: int = 0
    total_val_batches: int = 0
    global_step: int = 0
    estimated_stepping_batches: int = 0


class ValidateProgress(BaseModel):
    dataloader_idx: int = 0
    val_batch_idx: int = 0
    total_val_batches: int = 0


class TestProgress(BaseModel):
    dataloader_idx: int = 0
    test_batch_idx: int = 0
    total_test_batches: int = 0


class PredictProgress(BaseModel):
    dataloader_idx: int = 0
    predict_batch_idx: int = 0
    total_predict_batches: int = 0


class ProgressBarState(BaseModel):
    fit: FitProgress = Field(default_factory=FitProgress)
    val: ValidateProgress = Field(alias="validate", default_factory=ValidateProgress)
    test: TestProgress = Field(default_factory=TestProgress)
    predict: PredictProgress = Field(default_factory=PredictProgress)
    metrics: Dict[str, Union[float, str]] = {}


class TrainerState(BaseModel):
    fn: Optional[str] = None
    stage: Optional[str] = None



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The PanelFrontend wraps your Panel code in your LightningFlow."""
from __future__ import annotations

import inspect
import os
import pathlib
import subprocess
import sys
from typing import Callable, TextIO

from lightning.app.frontend.frontend import Frontend
from lightning.app.frontend.utils import _get_frontend_environment
from lightning.app.utilities.app_helpers import Logger
from lightning.app.utilities.cloud import is_running_in_cloud
from lightning.app.utilities.imports import requires
from lightning.app.utilities.log import get_logfile

_logger = Logger(__name__)


def _has_panel_autoreload() -> bool:
    """Returns True if the PANEL_AUTORELOAD environment variable is set to 'yes' or 'true'.

    Please note the casing of value does not matter
    """
    return os.environ.get("PANEL_AUTORELOAD", "no").lower() in ["yes", "y", "true"]


class PanelFrontend(Frontend):
    """The `PanelFrontend` enables you to serve Panel code as a Frontend for your LightningFlow.

    Reference: https://lightning.ai/lightning-docs/workflows/add_web_ui/panel/

    Args:
        entry_point: The path to a .py or .ipynb file, or a pure function. The file or function must contain your Panel
            code. The function can optionally accept an ``AppStateWatcher`` argument.

    Raises:
        TypeError: Raised if the ``entry_point`` provided is a class method

    Example:

    To use the `PanelFrontend`, you must first install the `panel` package:

    .. code-block:: bash

        pip install panel

    Create the files `panel_app_basic.py` and `app_basic.py` with the content below.

    **panel_app_basic.py**

    .. code-block:: python

        import panel as pn

        pn.panel("Hello **Panel ⚡** World").servable()

    **app_basic.py**

    .. code-block:: python

        import lightning as L
        from lightning.app.frontend.panel import PanelFrontend


        class LitPanel(L.LightningFlow):
            def configure_layout(self):
                return PanelFrontend("panel_app_basic.py")


        class LitApp(L.LightningFlow):
            def __init__(self):
                super().__init__()
                self.lit_panel = LitPanel()

            def configure_layout(self):
                return {"name": "home", "content": self.lit_panel}


        app = L.LightningApp(LitApp())

    Start the Lightning server with `lightning run app app_basic.py`.

    For development you can get Panel autoreload by setting the ``PANEL_AUTORELOAD``
    environment variable to 'yes', i.e. run
    ``PANEL_AUTORELOAD=yes lightning run app app_basic.py``
    """

    @requires("panel")
    def __init__(self, entry_point: str | Callable):
        super().__init__()

        if inspect.ismethod(entry_point):
            raise TypeError(
                "The `PanelFrontend` doesn't support `entry_point` being a method. Please, use a pure function."
            )

        self.entry_point = entry_point
        self._process: None | subprocess.Popen = None
        self._log_files: dict[str, TextIO] = {}
        _logger.debug("PanelFrontend Frontend with %s is initialized.", entry_point)

    def start_server(self, host: str, port: int, root_path: str = "") -> None:
        _logger.debug("PanelFrontend starting server on %s:%s", host, port)

        # 1: Prepare environment variables and arguments.
        env = _get_frontend_environment(
            self.flow.name,
            self.entry_point,
            port,
            host,
        )
        command = self._get_popen_args(host, port)

        if is_running_in_cloud():
            self._open_log_files()

        self._process = subprocess.Popen(command, env=env, **self._log_files)  # pylint: disable=consider-using-with

    def stop_server(self) -> None:
        if self._process is None:
            raise RuntimeError("Server is not running. Call `PanelFrontend.start_server()` first.")
        self._process.kill()
        self._close_log_files()

    def _close_log_files(self):
        for file_ in self._log_files.values():
            if not file_.closed:
                file_.close()
        self._log_files = {}

    def _open_log_files(self) -> None:
        # Don't log to file when developing locally. Makes it harder to debug.
        self._close_log_files()

        std_err_out = get_logfile("error.log")
        std_out_out = get_logfile("output.log")
        stderr = std_err_out.open("wb")
        stdout = std_out_out.open("wb")
        self._log_files = {"stdout": stderr, "stderr": stdout}

    def _get_popen_args(self, host: str, port: int) -> list:
        if callable(self.entry_point):
            path = str(pathlib.Path(__file__).parent / "panel_serve_render_fn.py")
        else:
            path = pathlib.Path(self.entry_point)

        abs_path = str(path)
        # The app is served at http://localhost:{port}/{flow}/{entry_point}
        # Lightning embeds http://localhost:{port}/{flow} but this redirects to the above and
        # seems to work fine.
        command = [
            sys.executable,
            "-m",
            "panel",
            "serve",
            abs_path,
            "--port",
            str(port),
            "--address",
            host,
            "--prefix",
            self.flow.name,
            "--allow-websocket-origin",
            _get_allowed_hosts(),
        ]
        if _has_panel_autoreload():
            command.append("--autoreload")
        _logger.debug("PanelFrontend command %s", command)
        return command


def _get_allowed_hosts() -> str:
    """Returns a comma separated list of host[:port] that should be allowed to connect."""
    # TODO: Enable only lightning.ai domain in the cloud
    return "*"



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import queue
import signal
import sys
import threading
import time
import traceback
import warnings
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass, field
from functools import partial
from threading import Event, Thread
from typing import Any, Callable, Dict, Generator, Optional, Set, Tuple, Type, TYPE_CHECKING, Union

from deepdiff import DeepDiff, Delta
from lightning_utilities.core.apply_func import apply_to_collection

from lightning.app.core import constants
from lightning.app.core.queues import MultiProcessQueue
from lightning.app.storage import Path
from lightning.app.storage.copier import _Copier, _copy_files
from lightning.app.storage.drive import _maybe_create_drive, Drive
from lightning.app.storage.path import _path_to_work_artifact
from lightning.app.storage.payload import Payload
from lightning.app.utilities.app_helpers import affiliation
from lightning.app.utilities.component import _set_work_context
from lightning.app.utilities.enum import (
    CacheCallsKeys,
    make_status,
    WorkFailureReasons,
    WorkStageStatus,
    WorkStopReasons,
)
from lightning.app.utilities.exceptions import CacheMissException, LightningSigtermStateException

if TYPE_CHECKING:
    from lightning.app import LightningWork
    from lightning.app.core.queues import BaseQueue

from lightning.app.utilities.app_helpers import Logger

logger = Logger(__name__)
_state_observer_lock = threading.Lock()


@dataclass
class Action:
    method: str = "run"
    args: Tuple = field(default_factory=lambda: ())
    kwargs: Dict = field(default_factory=lambda: {})


def unwrap(fn):
    if isinstance(fn, partial):
        fn = fn.keywords["work_run"]
    if isinstance(fn, ProxyWorkRun):
        fn = fn.work_run
    while hasattr(fn, "__wrapped__"):
        fn = fn.__wrapped__
    return fn


def _send_data_to_caller_queue(
    proxy, work: "LightningWork", caller_queue: "BaseQueue", data: Dict, call_hash: str
) -> Dict:

    proxy.has_sent = True

    if work._calls[CacheCallsKeys.LATEST_CALL_HASH] is None:
        work._calls[CacheCallsKeys.LATEST_CALL_HASH] = call_hash

    if call_hash not in work._calls:
        work._calls[call_hash] = {"statuses": []}
    else:
        # remove ret when relaunching the work.
        work._calls[call_hash].pop("ret", None)

    work._calls[call_hash]["statuses"].append(make_status(WorkStageStatus.PENDING))

    work_state = work.state

    # There is no need to send all call hashes to the work.
    calls = deepcopy(work_state["calls"])
    work_state["calls"] = {
        k: v for k, v in work_state["calls"].items() if k in (call_hash, CacheCallsKeys.LATEST_CALL_HASH)
    }

    data.update({"state": work_state})
    logger.debug(f"Sending to {work.name}: {data}")
    caller_queue.put(deepcopy(data))

    # Reset the calls entry.
    work_state["calls"] = calls
    work._restarting = False
    return work_state


@dataclass
class ProxyWorkRun:
    work_run: Callable
    work_name: str  # TODO: remove this argument and get the name from work.name directly
    work: "LightningWork"
    caller_queue: "BaseQueue"

    def __post_init__(self):
        self.work_state = None

    def __call__(self, *args, **kwargs):
        self.has_sent = False

        self._validate_call_args(args, kwargs)
        args, kwargs = self._process_call_args(args, kwargs)

        call_hash = self.work._call_hash(self.work_run, *self._convert_hashable(args, kwargs))
        entered = call_hash in self.work._calls
        returned = entered and "ret" in self.work._calls[call_hash]
        # TODO (tchaton): Handle spot instance retrieval differently from stopped work.
        stopped_on_sigterm = self.work._restarting and self.work.status.reason == WorkStopReasons.SIGTERM_SIGNAL_HANDLER

        data = {"args": args, "kwargs": kwargs, "call_hash": call_hash}

        # The if/else conditions are left un-compressed to simplify readability
        # for the readers.
        if self.work.cache_calls:
            if not entered or stopped_on_sigterm:
                _send_data_to_caller_queue(self, self.work, self.caller_queue, data, call_hash)
            else:
                if returned:
                    return
        else:
            if not entered or stopped_on_sigterm:
                _send_data_to_caller_queue(self, self.work, self.caller_queue, data, call_hash)
            else:
                if returned or stopped_on_sigterm:
                    # the previous task has completed and we can re-queue the next one.
                    # overriding the return value for next loop iteration.
                    _send_data_to_caller_queue(self, self.work, self.caller_queue, data, call_hash)
        if not self.work.parallel:
            raise CacheMissException("Task never called before. Triggered now")

    def _validate_call_args(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> None:
        """Validate the call args before they get passed to the run method of the Work.

        Currently, this performs a check against strings that look like filesystem paths and may need to be wrapped with
        a Lightning Path by the user.
        """

        def warn_if_pathlike(obj: Union[os.PathLike, str]):
            if isinstance(obj, Path):
                return
            if os.sep in str(obj) and os.path.exists(obj):
                # NOTE: The existence check is wrong in general, as the file will never exist on the disk
                # where the flow is running unless we are running locally
                warnings.warn(
                    f"You passed a the value {obj!r} as an argument to the `run()` method of {self.work_name} and"
                    f" it looks like this is a path to a file or a folder. Consider wrapping this path in a"
                    f" `lightning.app.storage.Path` object to be able to access these files in your Work.",
                    UserWarning,
                )

        apply_to_collection(args, dtype=(os.PathLike, str), function=warn_if_pathlike)
        apply_to_collection(kwargs, dtype=(os.PathLike, str), function=warn_if_pathlike)

    @staticmethod
    def _process_call_args(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """Processes all positional and keyword arguments before they get passed to the caller queue and sent to
        the LightningWork.

        Currently, this method only applies sanitization to Lightning Path objects.

        Args:
            args: The tuple of positional arguments passed to the run method.
            kwargs: The dictionary of named arguments passed to the run method.

        Returns:
            The positional and keyword arguments in the same order they were passed in.
        """

        def sanitize(obj: Union[Path, Drive]) -> Union[Path, Dict]:
            if isinstance(obj, Path):
                # create a copy of the Path and erase the consumer
                # the LightningWork on the receiving end of the caller queue will become the new consumer
                # this is necessary to make the Path deepdiff-hashable
                path_copy = Path(obj)
                path_copy._sanitize()
                path_copy._consumer = None
                return path_copy
            return obj.to_dict()

        return apply_to_collection((args, kwargs), dtype=(Path, Drive), function=sanitize)

    @staticmethod
    def _convert_hashable(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """Processes all positional and keyword arguments before they get passed to the caller queue and sent to
        the LightningWork.

        Currently, this method only applies sanitization to Hashable Objects.

        Args:
            args: The tuple of positional arguments passed to the run method.
            kwargs: The dictionary of named arguments passed to the run method.

        Returns:
            The positional and keyword arguments in the same order they were passed in.
        """
        from lightning.app.utilities.types import Hashable

        def sanitize(obj: Hashable) -> Union[Path, Dict]:
            return obj.to_dict()

        return apply_to_collection((args, kwargs), dtype=Hashable, function=sanitize)


class WorkStateObserver(Thread):
    """This thread runs alongside LightningWork and periodically checks for state changes. If the state changed
    from one interval to the next, it will compute the delta and add it to the queue which is connected to the
    Flow. This enables state changes to be captured that are not triggered through a setattr call.

    Args:
        work: The LightningWork for which the state should be monitored
        delta_queue: The queue to send deltas to when state changes occur
        interval: The interval at which to check for state changes.

    Example:

        class Work(LightningWork):
            ...

            def run(self):
                # This update gets sent to the Flow once the thread compares the new state with the previous one
                self.list.append(1)
    """

    def __init__(
        self,
        work: "LightningWork",
        delta_queue: "BaseQueue",
        flow_to_work_delta_queue: Optional["BaseQueue"] = None,
        error_queue: Optional["BaseQueue"] = None,
        interval: float = 1,
    ) -> None:
        super().__init__(daemon=True)
        self.started = False
        self._work = work
        self._delta_queue = delta_queue
        self._flow_to_work_delta_queue = flow_to_work_delta_queue
        self._error_queue = error_queue
        self._interval = interval
        self._exit_event = Event()
        self._delta_memory = []
        self._last_state = deepcopy(self._work.state)

    def run(self) -> None:
        self.started = True
        while not self._exit_event.is_set():
            time.sleep(self._interval)
            # Run the thread only if active
            self.run_once()

    @staticmethod
    def get_state_changed_from_queue(q: "BaseQueue", timeout: Optional[int] = None):
        try:
            delta = q.get(timeout=timeout or q.default_timeout)
            return delta
        except queue.Empty:
            return None

    def run_once(self) -> None:
        with _state_observer_lock:
            # Add all deltas the LightningWorkSetAttrProxy has processed and sent to the Flow already while
            # the WorkStateObserver was sleeping
            for delta in self._delta_memory:
                self._last_state += delta
            self._delta_memory.clear()

            # The remaining delta is the result of state updates triggered outside the setattr, e.g, by a list append
            delta = Delta(DeepDiff(self._last_state, self._work.state, verbose_level=2))
            if not delta.to_dict():
                return
            self._last_state = deepcopy(self._work.state)
            self._delta_queue.put(ComponentDelta(id=self._work.name, delta=delta))

        if self._flow_to_work_delta_queue:
            while True:
                deep_diff = self.get_state_changed_from_queue(self._flow_to_work_delta_queue)
                if not isinstance(deep_diff, dict):
                    break
                try:
                    with _state_observer_lock:
                        self._work.apply_flow_delta(Delta(deep_diff, raise_errors=True))
                except Exception as e:
                    print(traceback.print_exc())
                    self._error_queue.put(e)
                    raise e

    def join(self, timeout: Optional[float] = None) -> None:
        self._exit_event.set()
        super().join(timeout)


@dataclass
class LightningWorkSetAttrProxy:
    """This wrapper around the ``LightningWork.__setattr__`` ensures that state changes get sent to the delta queue
    to be reflected in the Flow.

    Example:

        class Work(LightningWork):
            ...

            def run(self):
                self.var += 1  # This update gets sent to the Flow immediately
    """

    work_name: str
    work: "LightningWork"
    delta_queue: "BaseQueue"
    state_observer: Optional["WorkStateObserver"]

    def __call__(self, name: str, value: Any) -> None:
        logger.debug(f"Setting {name}: {value}")
        with _state_observer_lock:
            state = deepcopy(self.work.state)
            self.work._default_setattr(name, value)
            delta = Delta(DeepDiff(state, self.work.state, verbose_level=2))
            if not delta.to_dict():
                return

            # push the delta only if there is any
            self.delta_queue.put(ComponentDelta(id=self.work_name, delta=delta))

            # add the delta to the buffer to let WorkStateObserver know we already sent this one to the Flow
            if self.state_observer:
                self.state_observer._delta_memory.append(delta)


@dataclass
class ComponentDelta:
    id: str
    delta: Delta


@dataclass
class WorkRunExecutor:
    work: "LightningWork"
    work_run: Callable
    delta_queue: "BaseQueue"
    enable_start_observer: bool = True

    def __call__(self, *args, **kwargs):
        return self.work_run(*args, **kwargs)

    @contextmanager
    def enable_spawn(self) -> Generator:
        self.work._setattr_replacement = None
        self.work._backend = None
        self._clean_queues()
        yield

    def _clean_queues(self):
        if not isinstance(self.work._request_queue, MultiProcessQueue):
            self.work._request_queue = self.work._request_queue.to_dict()
            self.work._response_queue = self.work._response_queue.to_dict()

    @staticmethod
    def process_queue(queue):
        from lightning.app.core.queues import HTTPQueue, RedisQueue

        if isinstance(queue, dict):
            queue_type = queue.pop("type")
            if queue_type == "redis":
                return RedisQueue.from_dict(queue)
            else:
                return HTTPQueue.from_dict(queue)
        return queue


@dataclass
class WorkRunner:
    work: "LightningWork"
    work_name: str
    caller_queue: "BaseQueue"
    delta_queue: "BaseQueue"
    readiness_queue: "BaseQueue"
    error_queue: "BaseQueue"
    request_queue: "BaseQueue"
    response_queue: "BaseQueue"
    copy_request_queue: "BaseQueue"
    copy_response_queue: "BaseQueue"
    flow_to_work_delta_queue: Optional["BaseQueue"] = None
    run_executor_cls: Type[WorkRunExecutor] = WorkRunExecutor

    def __post_init__(self):
        self.parallel = self.work.parallel
        self.copier: Optional[_Copier] = None
        self.state_observer: Optional[WorkStateObserver] = None

    def __call__(self):
        self.setup()
        while True:
            try:
                self.run_once()
            except KeyboardInterrupt:
                if self.state_observer:
                    if self.state_observer.started:
                        self.state_observer.join(0)
                    self.state_observer = None
                self.copier.join(0)
            except LightningSigtermStateException as e:
                logger.debug("Exiting")
                os._exit(e.exit_code)
            except Exception as e:
                # Inform the flow the work failed. This would fail the entire application.
                self.error_queue.put(e)
                # Terminate the threads
                if self.state_observer:
                    if self.state_observer.started:
                        self.state_observer.join(0)
                    self.state_observer = None
                self.copier.join(0)
                raise e

    def setup(self):
        from lightning.app.utilities.state import AppState

        _set_work_context()

        # 1. Make the AppState aware of the affiliation of the work.
        # hacky: attach affiliation to be know from the AppState object
        AppState._MY_AFFILIATION = affiliation(self.work)

        # 2. Attach the queues to the work.
        # Each work gets their own request- and response storage queue for communicating with the storage orchestrator
        self.work._request_queue = self.request_queue
        self.work._response_queue = self.response_queue

        # 3. Starts the Copier thread. This thread enables transfering files using
        # the Path object between works.
        self.copier = _Copier(self.work, self.copy_request_queue, self.copy_response_queue)
        self.copier.setDaemon(True)
        self.copier.start()

        # 4. If the work is restarting, reload the latest state.
        # TODO (tchaton) Add support for capturing the latest state.
        if self.work._restarting:
            self.work.load_state_dict(self.work.state)

        # 5. Inform the flow that the work is ready to receive data through the caller queue.
        self.readiness_queue.put(True)

    def run_once(self):
        # 1. Wait for the caller queue data.
        called: Dict[str, Any] = self.caller_queue.get()
        logger.debug(f"Work {self.work_name} {called}")

        # 2. Extract the info from the caller queue data and process the input arguments. Arguments can contain
        # Lightning Path objects; if they don't have a consumer, the current Work will become one.
        call_hash = called["call_hash"]
        args, kwargs = self._process_call_args(called["args"], called["kwargs"])

        # 3. Register the signal handler for spot instances.
        # `SIGUSR1` signal isn't supported on windows.
        # TODO (tchaton) Add support for windows
        if sys.platform != "win32":
            signal.signal(signal.SIGTERM, partial(self._sigterm_signal_handler, call_hash=call_hash))

        # 4. Set the received state to the work.
        self.work.set_state(called["state"])

        # 5. Transfer all paths in the state automatically if they have an origin and exist
        self._transfer_path_attributes()

        # 6. Create the state observer thread.
        if self.run_executor_cls.enable_start_observer:
            self.state_observer = WorkStateObserver(
                self.work,
                delta_queue=self.delta_queue,
                flow_to_work_delta_queue=self.flow_to_work_delta_queue,
                error_queue=self.error_queue,
            )

        # 7. Deepcopy the work state and send the first `RUNNING` status delta to the flow.
        reference_state = deepcopy(self.work.state)

        # Set the internal IP address.
        # Set this here after the state observer is initialized, since it needs to record it as a change and send
        # it back to the flow
        default_internal_ip = "127.0.0.1" if constants.LIGHTNING_CLOUDSPACE_HOST is None else "0.0.0.0"
        self.work._internal_ip = os.environ.get("LIGHTNING_NODE_IP", default_internal_ip)

        # 8. Patch the setattr method of the work. This needs to be done after step 4, so we don't
        # send delta while calling `set_state`.
        self._proxy_setattr()

        if self._is_starting(called, reference_state, call_hash):
            return

        # 9. Inform the flow the work is running and add the delta to the deepcopy.
        self.work._calls[CacheCallsKeys.LATEST_CALL_HASH] = call_hash
        self.work._calls[call_hash]["statuses"].append(make_status(WorkStageStatus.RUNNING))
        delta = Delta(DeepDiff(reference_state, self.work.state))
        self.delta_queue.put(ComponentDelta(id=self.work_name, delta=delta))

        # 10. Unwrap the run method if wrapped.
        work_run = self.work.run
        if hasattr(work_run, "__wrapped__"):
            work_run = work_run.__wrapped__

        # 11. Start the state observer thread. It will look for state changes and send them back to the Flow
        # The observer has to be initialized here, after the set_state call above so that the thread can start with
        # the proper initial state of the work
        if self.run_executor_cls.enable_start_observer:
            self.state_observer.start()

        # 12. Run the `work_run` method.
        # If an exception is raised, send a `FAILED` status delta to the flow and call the `on_exception` hook.
        try:
            ret = self.run_executor_cls(self.work, work_run, self.delta_queue)(*args, **kwargs)
        except LightningSigtermStateException as e:
            raise e
        except BaseException as e:
            # 10.2 Send failed delta to the flow.
            reference_state = deepcopy(self.work.state)
            exp, val, tb = sys.exc_info()
            listing = traceback.format_exception(exp, val, tb)
            user_exception = False
            used_runpy = False
            trace = []
            for p in listing:
                if "runpy.py" in p:
                    trace = []
                    used_runpy = True
                if user_exception:
                    trace.append(p)
                if "ret = self.run_executor_cls(" in p:
                    user_exception = True

            if used_runpy:
                trace = trace[1:]

            self.work._calls[call_hash]["statuses"].append(
                make_status(
                    WorkStageStatus.FAILED,
                    message=str("\n".join(trace)),
                    reason=WorkFailureReasons.USER_EXCEPTION,
                )
            )
            self.delta_queue.put(
                ComponentDelta(
                    id=self.work_name, delta=Delta(DeepDiff(reference_state, self.work.state, verbose_level=2))
                )
            )
            self.work.on_exception(e)
            print("########## CAPTURED EXCEPTION ###########")
            print(traceback.print_exc())
            print("########## CAPTURED EXCEPTION ###########")
            return

        # 13. Destroy the state observer.
        if self.run_executor_cls.enable_start_observer:
            if self.state_observer.started:
                self.state_observer.join(0)
        self.state_observer = None

        # 14. Copy all artifacts to the shared storage so other Works can access them while this Work gets scaled down
        persist_artifacts(work=self.work)

        # 15. An asynchronous work shouldn't return a return value.
        if ret is not None:
            raise RuntimeError(
                f"Your work {self.work} shouldn't have a return value. Found {ret}."
                "HINT: Use the Payload API instead."
            )

        # 17. DeepCopy the state and send the latest delta to the flow.
        # use the latest state as we have already sent delta
        # during its execution.
        # inform the task has completed
        reference_state = deepcopy(self.work.state)
        self.work._calls[call_hash]["statuses"].append(make_status(WorkStageStatus.SUCCEEDED))
        self.work._calls[call_hash]["ret"] = ret
        self.delta_queue.put(
            ComponentDelta(id=self.work_name, delta=Delta(DeepDiff(reference_state, self.work.state, verbose_level=2)))
        )

        # 18. Update the work for the next delta if any.
        self._proxy_setattr(cleanup=True)

    def _sigterm_signal_handler(self, signum, frame, call_hash: str) -> None:
        """Signal handler used to react when spot instances are being retrived."""
        logger.info(f"Received SIGTERM signal. Gracefully terminating {self.work.name.replace('root.', '')}...")
        persist_artifacts(work=self.work)
        with _state_observer_lock:
            self.work.on_exit()
            self.work._calls[call_hash]["statuses"] = []
            state = deepcopy(self.work.state)
            self.work._calls[call_hash]["statuses"].append(
                make_status(WorkStageStatus.STOPPED, reason=WorkStopReasons.SIGTERM_SIGNAL_HANDLER)
            )

        # kill the thread as the job is going to be terminated.
        if self.state_observer:
            if self.state_observer.started:
                self.state_observer.join(0)
            self.state_observer = None
        delta = Delta(DeepDiff(state, deepcopy(self.work.state), verbose_level=2))
        self.delta_queue.put(ComponentDelta(id=self.work_name, delta=delta))

        self.copier.join(0)
        raise LightningSigtermStateException(0)

    def _proxy_setattr(self, cleanup: bool = False):
        _proxy_setattr(self.work, self.delta_queue, self.state_observer, cleanup=cleanup)

    def _process_call_args(
        self, args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """Process the arguments that were passed in to the ``run()`` method of the
        :class:`lightning.app.core.work.LightningWork`.

        This method currently only implements special treatments for the :class:`lightning.app.storage.path.Path`
        objects. Any Path objects that get passed into the run method get attached to the Work automatically, i.e.,
        the Work becomes the `origin` or the `consumer` if they were not already before. Additionally,
        if the file or folder under the Path exists, we transfer it.

        Args:
            args: The tuple of positional arguments passed to the run method.
            kwargs: The dictionary of named arguments passed to the run method.

        Returns:
            The positional and keyword arguments in the same order they were passed in.
        """

        def _attach_work_and_get(transporter: Union[Path, Payload, dict]) -> Union[Path, Drive, dict, Any]:
            if not transporter.origin_name:
                # If path/payload is not attached to an origin, there is no need to attach or transfer anything
                return transporter

            transporter._attach_work(self.work)
            transporter._attach_queues(self.work._request_queue, self.work._response_queue)
            if transporter.exists_remote():
                # All paths/payloads passed to the `run` method under a Lightning obj need to be copied (if they exist)
                if isinstance(transporter, Payload):
                    transporter.get()
                else:
                    transporter.get(overwrite=True)
            return transporter

        def _handle_drive(dict):
            return _maybe_create_drive(self.work_name, dict)

        args, kwargs = apply_to_collection((args, kwargs), dtype=(Path, Payload), function=_attach_work_and_get)
        return apply_to_collection((args, kwargs), dtype=dict, function=_handle_drive)

    def _transfer_path_attributes(self) -> None:
        """Transfer all Path attributes in the Work if they have an origin and exist."""
        for name in self.work._paths:
            path = getattr(self.work, name)
            if isinstance(path, str):
                path = Path(path)
                path._attach_work(self.work)
            if path.origin_name and path.origin_name != self.work.name and path.exists_remote():
                path.get(overwrite=True)

    def _is_starting(self, called, reference_state, call_hash) -> bool:
        if len(called["args"]) == 1 and isinstance(called["args"][0], Action):
            action = called["args"][0]
            if action.method == "start":
                # 9. Inform the flow the work is running and add the delta to the deepcopy.
                self.work._calls[CacheCallsKeys.LATEST_CALL_HASH] = call_hash
                self.work._calls[call_hash]["statuses"].append(make_status(WorkStageStatus.STARTED))
                delta = Delta(DeepDiff(reference_state, self.work.state))
                self.delta_queue.put(ComponentDelta(id=self.work_name, delta=delta))
                self._proxy_setattr(cleanup=True)
                return True
            else:
                raise Exception("Only the `start` action is supported right now !")
        return False


def persist_artifacts(work: "LightningWork") -> None:
    """Copies all :class:`~lightning.app.storage.path.Path` referenced by the given LightningWork to the shared
    storage.

    Files that don't exist or do not originate from the given Work will be skipped.
    """
    artifact_paths = [getattr(work, name) for name in work._paths]
    # only copy files that belong to this Work, i.e., when the path's origin refers to the current Work
    artifact_paths = [path for path in artifact_paths if isinstance(path, Path) and path.origin_name == work.name]

    for name in work._state:
        if isinstance(getattr(work, name), Payload):
            artifact_path = pathlib.Path(name).resolve()
            payload = getattr(work, name)
            payload.save(payload.value, artifact_path)
            artifact_paths.append(artifact_path)

    missing_artifacts: Set[str] = set()
    destination_paths = []
    for artifact_path in artifact_paths:
        artifact_path = pathlib.Path(artifact_path).absolute()
        if not artifact_path.exists():
            missing_artifacts.add(str(artifact_path))
            continue
        destination_path = _path_to_work_artifact(artifact_path, work)
        _copy_files(artifact_path, destination_path)
        destination_paths.append(destination_path)

    if missing_artifacts:
        warnings.warn(
            f"{len(missing_artifacts)} artifacts could not be saved because they don't exist:"
            f" {','.join(missing_artifacts)}.",
            UserWarning,
        )
    else:
        logger.debug(
            f"All {destination_paths} artifacts from Work {work.name} successfully "
            "stored at {artifacts_path(work.name)}."
        )


def _proxy_setattr(work, delta_queue, state_observer: Optional[WorkStateObserver], cleanup: bool = False):
    if cleanup:
        setattr_proxy = None
    else:
        setattr_proxy = LightningWorkSetAttrProxy(
            work.name,
            work,
            delta_queue=delta_queue,
            state_observer=state_observer,
        )
    work._setattr_replacement = setattr_proxy



from lightning.app.core.app import LightningApp
from lightning.app.core.flow import LightningFlow
from lightning.app.core.work import LightningWork

__all__ = ["LightningApp", "LightningFlow", "LightningWork"]



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as t

from typing_extensions import Protocol, runtime_checkable

from lightning.app import LightningFlow, LightningWork
from lightning.app.structures import Dict, List

Component = t.Union[LightningFlow, LightningWork, Dict, List]
ComponentTuple = (LightningFlow, LightningWork, Dict, List)


@runtime_checkable
class Hashable(Protocol):
    def to_dict(self) -> t.Dict[str, t.Any]:
        ...



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
from typing import Any, Optional, Type

import lightning.pytorch as pl
from lightning.pytorch.accelerators import CUDAAccelerator
from lightning.pytorch.loops import Loop
from lightning.pytorch.loops.epoch import TrainingEpochLoop
from lightning.pytorch.loops.epoch.training_epoch_loop import _OUTPUTS_TYPE as _EPOCH_OUTPUTS_TYPE
from lightning.pytorch.loops.utilities import _is_max_limit_reached, _set_sampler_epoch
from lightning.pytorch.trainer.connectors.logger_connector.result import _ResultCollection
from lightning.pytorch.trainer.progress import Progress
from lightning.pytorch.trainer.supporters import CombinedLoader, TensorRunningAccum
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.fetching import (
    AbstractDataFetcher,
    DataFetcher,
    DataLoaderIterDataFetcher,
    InterBatchParallelDataFetcher,
)
from lightning.pytorch.utilities.model_helpers import is_overridden
from lightning.pytorch.utilities.rank_zero import rank_zero_debug, rank_zero_info, rank_zero_warn
from lightning.pytorch.utilities.signature_utils import is_param_in_hook_signature

log = logging.getLogger(__name__)


class FitLoop(Loop[None]):
    """This Loop iterates over the epochs to run the training.

    Args:
        min_epochs: The minimum number of epochs
        max_epochs: The maximum number of epochs, can be set -1 to turn this limit off
    """

    def __init__(
        self,
        min_epochs: Optional[int] = 0,
        max_epochs: Optional[int] = None,
    ) -> None:
        super().__init__()
        if isinstance(max_epochs, int) and max_epochs < -1:
            # Allow max_epochs to be zero, since this will be handled by fit_loop.done
            raise MisconfigurationException(
                f"`max_epochs` must be a non-negative integer or -1. You passed in {max_epochs}."
            )

        self.max_epochs = max_epochs
        self.min_epochs = min_epochs
        self.epoch_loop = TrainingEpochLoop()
        self.epoch_progress = Progress()

        self._is_fresh_start_epoch: bool = True
        self._outputs: _EPOCH_OUTPUTS_TYPE = []
        self._data_fetcher: Optional[AbstractDataFetcher] = None

    @property
    def total_batch_idx(self) -> int:
        """Returns the current batch index (across epochs)"""
        return self.epoch_loop.total_batch_idx

    @property
    def batch_idx(self) -> int:
        """Returns the current batch index (within this epoch)"""
        return self.epoch_loop.batch_idx

    @property
    def split_idx(self) -> int:
        """Returns the index of the current batch split (within the current batch) for bptt."""
        return self.epoch_loop.batch_loop.split_idx

    @property
    def min_steps(self) -> Optional[int]:
        # TODO(@justusschock): Why aren't we using the attribute in this class?
        """Returns the minimum number of steps to run."""
        return self.epoch_loop.min_steps

    @min_steps.setter
    def min_steps(self, value: Optional[int]) -> None:
        """Sets the minimum number of steps (forwards to epoch_loop)"""
        # TODO: This setter is required by debugging connector (fast dev run), should be avoided
        self.epoch_loop.min_steps = value

    @property
    def max_steps(self) -> int:
        """Returns the maximum number of steps to run."""
        return self.epoch_loop.max_steps

    @max_steps.setter
    def max_steps(self, value: int) -> None:
        """Sets the maximum number of steps (forwards to epoch_loop)"""
        # TODO: This setter is required by debugging connector (fast dev run), should be avoided
        if value < -1:
            raise MisconfigurationException(
                f"`max_steps` must be a non-negative integer or -1 (infinite steps). You passed in {value}."
            )
        self.epoch_loop.max_steps = value

    @property
    def running_loss(self) -> TensorRunningAccum:
        """Returns the running loss."""
        return self.epoch_loop.batch_loop.running_loss

    @Loop.restarting.setter
    def restarting(self, restarting: bool) -> None:
        # if the last epoch completely finished, we are not actually restarting
        values = self.epoch_progress.current.ready, self.epoch_progress.current.started
        epoch_unfinished = any(v != self.epoch_progress.current.processed for v in values)
        restarting = restarting and epoch_unfinished or self._iteration_based_training()
        Loop.restarting.fset(self, restarting)  # call the parent setter

    @property
    def prefetch_batches(self) -> int:
        is_unsized = self.trainer.num_training_batches == float("inf")
        inter_batch_parallelism = os.getenv("PL_INTER_BATCH_PARALLELISM", "0") == "1"
        return 1 if is_unsized or inter_batch_parallelism else 0

    @property
    def _skip_backward(self) -> bool:
        """Determines whether the loop will skip backward during automatic optimization."""
        return self.epoch_loop.batch_loop.optimizer_loop._skip_backward

    @_skip_backward.setter
    def _skip_backward(self, value: bool) -> None:
        """Determines whether the loop will skip backward during automatic optimization."""
        self.epoch_loop.batch_loop.optimizer_loop._skip_backward = value

    @property
    def _results(self) -> _ResultCollection:
        if self.trainer.training:
            return self.epoch_loop._results
        if self.trainer.validating:
            return self.epoch_loop.val_loop._results
        raise RuntimeError("`FitLoop._results` property isn't defined. Accessed outside of scope")

    @property
    def _can_stop_early(self) -> bool:
        met_min_epochs = self.epoch_progress.current.processed >= self.min_epochs if self.min_epochs else True
        met_min_steps = self.epoch_loop.global_step >= self.min_steps if self.min_steps else True
        return met_min_epochs and met_min_steps

    @property
    def done(self) -> bool:
        """Evaluates when to leave the loop."""
        if self.trainer.num_training_batches == 0:
            rank_zero_info("`Trainer.fit` stopped: No training batches.")
            return True

        # TODO: Move track steps inside training loop and move part of these condition inside training loop
        stop_steps = _is_max_limit_reached(self.epoch_loop.global_step, self.max_steps)
        if stop_steps:
            rank_zero_info(f"`Trainer.fit` stopped: `max_steps={self.max_steps!r}` reached.")
            return True

        # `processed` is increased before `on_train_epoch_end`, the hook where checkpoints are typically saved.
        # we use it here because the checkpoint data won't have `completed` increased yet
        assert isinstance(self.max_epochs, int)
        stop_epochs = _is_max_limit_reached(self.epoch_progress.current.processed, self.max_epochs)
        if stop_epochs:
            # in case they are not equal, override so `trainer.current_epoch` has the expected value
            self.epoch_progress.current.completed = self.epoch_progress.current.processed
            rank_zero_info(f"`Trainer.fit` stopped: `max_epochs={self.max_epochs!r}` reached.")
            return True

        if self.trainer.should_stop and self._can_stop_early:
            rank_zero_debug("`Trainer.fit` stopped: `trainer.should_stop` was set.")
            return True

        return False

    @property
    def skip(self) -> bool:
        """Whether we should skip the training and immediately return from the call to :meth:`run`."""
        # since `trainer.num_training_batches` depends on the `train_dataloader` but that won't be called
        # until `on_run_start`, we use `limit_train_batches` instead
        return self.done or self.trainer.limit_train_batches == 0

    def connect(self, epoch_loop: TrainingEpochLoop) -> None:  # type: ignore[override]
        """Connects a training epoch loop to this fit loop."""
        self.epoch_loop = epoch_loop

    def reset(self) -> None:
        """Resets the internal state of this loop."""
        if self.restarting:
            self.epoch_progress.reset_on_restart()

    def on_run_start(self) -> None:
        """Calls the ``on_train_start`` hook."""
        # update the current_epoch in-case of checkpoint reload
        if not self._iteration_based_training():
            self.epoch_progress.current.completed = self.epoch_progress.current.processed

        self.trainer.reset_train_dataloader(self.trainer.lightning_module)
        # reload the evaluation dataloaders too for proper display in the progress bar
        if self.epoch_loop._should_check_val_epoch():
            self.epoch_loop.val_loop._reload_evaluation_dataloaders()

        data_fetcher_cls = _select_data_fetcher(self.trainer)
        self._data_fetcher = data_fetcher_cls(prefetch_batches=self.prefetch_batches)

        self._is_fresh_start_epoch = True
        self._results.to(device=self.trainer.lightning_module.device)

        self.trainer._call_callback_hooks("on_train_start")
        self.trainer._call_lightning_module_hook("on_train_start")
        self.trainer._call_strategy_hook("on_train_start")

    def on_advance_start(self) -> None:
        """Prepares the dataloader for training and calls the hook ``on_train_epoch_start``"""
        model = self.trainer.lightning_module

        # reset train dataloader
        if not self._is_fresh_start_epoch and self.trainer._data_connector._should_reload_train_dl:
            log.detail(f"{self.__class__.__name__}: resetting train dataloader")
            self.trainer.reset_train_dataloader(model)
        self._is_fresh_start_epoch = False

        # reset outputs here instead of in `reset` as they are not accumulated between epochs
        self._outputs = []

        if self.trainer.train_dataloader is not None:
            assert isinstance(self.trainer.train_dataloader, CombinedLoader)
            _set_sampler_epoch(self.trainer.train_dataloader, self.epoch_progress.current.processed)

        # changing gradient according accumulation_scheduler
        self.trainer.accumulation_scheduler.on_train_epoch_start(self.trainer, self.trainer.lightning_module)

        # stores accumulated grad fractions per batch
        self.epoch_loop.batch_loop.accumulated_loss.reset(window_length=self.trainer.accumulate_grad_batches)

        self.epoch_progress.increment_ready()

        self.trainer._logger_connector.on_epoch_start()

        self.trainer._call_callback_hooks("on_train_epoch_start")
        self.trainer._call_lightning_module_hook("on_train_epoch_start")

        self.epoch_progress.increment_started()

    def advance(self) -> None:
        """Runs one whole epoch."""
        log.detail(f"{self.__class__.__name__}: advancing loop")
        assert self.trainer.train_dataloader is not None
        dataloader = self.trainer.train_dataloader

        def batch_to_device(batch: Any) -> Any:
            batch = self.trainer.lightning_module._on_before_batch_transfer(batch, dataloader_idx=0)
            batch = self.trainer._call_strategy_hook("batch_to_device", batch, dataloader_idx=0)
            return batch

        assert self._data_fetcher is not None
        self._data_fetcher.setup(dataloader, batch_to_device=batch_to_device)
        with self.trainer.profiler.profile("run_training_epoch"):
            self._outputs = self.epoch_loop.run(self._data_fetcher)

    def on_advance_end(self) -> None:
        # inform logger the batch loop has finished
        self.trainer._logger_connector.epoch_end_reached()

        # get the model and call model.training_epoch_end
        model = self.trainer.lightning_module
        if is_overridden("training_epoch_end", model) and self._outputs:
            epoch_end_outputs = self.epoch_loop._prepare_outputs_training_epoch_end(
                self._outputs,
                lightning_module=model,
                num_optimizers=len(self.trainer.optimizers),
            )
            # run lightning module hook training_epoch_end
            # refresh the result for custom logging at the epoch level
            epoch_end_outputs = self.trainer._call_lightning_module_hook("training_epoch_end", epoch_end_outputs)
            if epoch_end_outputs is not None:
                raise MisconfigurationException(
                    "`training_epoch_end` expects a return of None. "
                    "HINT: remove the return statement in `training_epoch_end`."
                )
        # free memory
        self._outputs = []

        self.epoch_progress.increment_processed()

        # call train epoch end hooks
        self.trainer._call_callback_hooks("on_train_epoch_end")
        self.trainer._call_lightning_module_hook("on_train_epoch_end")

        self.trainer._logger_connector.on_epoch_end()

        if self.epoch_loop._num_ready_batches_reached():
            # if we are restarting and the above condition holds, it's because we are reloading an epoch-end checkpoint.
            # since metric-based schedulers require access to metrics and those are not currently saved in the
            # checkpoint, the plateau schedulers shouldn't be updated
            self.epoch_loop.update_lr_schedulers("epoch", update_plateau_schedulers=not self.restarting)

        # we manually decrease here because loggers expect that the same step is used when logging epoch-end metrics
        # even when the batch loop has finished
        self.epoch_loop._batches_that_stepped -= 1
        # log epoch metrics
        self.trainer._logger_connector.update_train_epoch_metrics()
        self.epoch_loop._batches_that_stepped += 1

        self.epoch_progress.increment_completed()

        # if fault tolerant is enabled and process has been notified, exit.
        self.trainer._exit_gracefully_on_signal()

    def on_run_end(self) -> None:
        """Calls the ``on_train_end`` hook."""
        log.detail(f"{self.__class__.__name__}: train run ended")

        # hook
        self.trainer._call_callback_hooks("on_train_end")
        self.trainer._call_lightning_module_hook("on_train_end")
        self.trainer._call_strategy_hook("on_train_end")

    def teardown(self) -> None:
        if self._data_fetcher is not None:
            self._data_fetcher.teardown()
            self._data_fetcher = None
        self.epoch_loop.teardown()

    def _should_accumulate(self) -> bool:
        """Whether the gradients should be accumulated."""
        return self.epoch_loop._should_accumulate()

    def _iteration_based_training(self) -> bool:
        return self.trainer.max_steps != -1


def _select_data_fetcher(trainer: "pl.Trainer") -> Type[AbstractDataFetcher]:
    training_step_fx = getattr(trainer.lightning_module, "training_step")
    if is_param_in_hook_signature(training_step_fx, "dataloader_iter", explicit=True):
        rank_zero_warn(
            "Found `dataloader_iter` argument in the `training_step`. Note that the support for "
            "this signature is experimental and the behavior is subject to change."
        )
        return DataLoaderIterDataFetcher
    elif os.getenv("PL_INTER_BATCH_PARALLELISM", "0") == "1":
        if not isinstance(trainer.accelerator, CUDAAccelerator):
            raise MisconfigurationException("Inter batch parallelism is available only when using Nvidia GPUs.")
        return InterBatchParallelDataFetcher
    return DataFetcher



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import pathlib
import pickle
from abc import ABC, abstractmethod
from time import sleep
from typing import Any, Optional, TYPE_CHECKING, Union

from lightning.app.core.constants import REMOTE_STORAGE_WAIT
from lightning.app.core.queues import BaseQueue
from lightning.app.storage.path import _filesystem, _shared_storage_path, Path
from lightning.app.storage.requests import _ExistsRequest, _ExistsResponse, _GetRequest, _GetResponse
from lightning.app.utilities.app_helpers import Logger
from lightning.app.utilities.component import _is_flow_context

_logger = Logger(__name__)

if TYPE_CHECKING:
    from lightning.app.core.work import LightningWork


class _BasePayload(ABC):
    def __init__(self, value: Any) -> None:
        self._value = value
        # the attribute name given to the payload
        self._name: Optional[str] = None
        # the origin is the work that created this Path and wants to expose file(s)
        self._origin: Optional[Union["LightningWork", str]] = None
        # the consumer is the Work that needs access to the file(s) from the consumer
        self._consumer: Optional[Union["LightningWork", str]] = None
        self._metadata = {}
        # request queue: used to transfer message to storage orchestrator
        self._request_queue: Optional[BaseQueue] = None
        # response queue: used to receive status message from storage orchestrator
        self._response_queue: Optional[BaseQueue] = None

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def value(self) -> Optional[Any]:
        """The real object that this payload holds."""
        return self._value

    @property
    def hash(self) -> Optional[str]:
        """The hash of this Payload uniquely identifies the payload and the associated origin Work.

        Returns ``None`` if the origin is not defined, i.e., this Path did not yet get attached to a LightningWork.
        """
        if self._origin is None:
            return None
        contents = f"{self.origin_name}/{self.consumer_name}/{self.name}"
        return hashlib.sha1(contents.encode("utf-8")).hexdigest()

    @property
    def origin_name(self) -> str:
        """The name of the LightningWork where this payload was first created.

        Attaching a Payload to a LightningWork will automatically make it the `origin`.
        """
        from lightning.app.core.work import LightningWork

        return self._origin.name if isinstance(self._origin, LightningWork) else self._origin

    @property
    def consumer_name(self) -> str:
        """The name of the LightningWork where this payload is being accessed.

        By default, this is the same as the :attr:`origin_name`.
        """
        from lightning.app.core.work import LightningWork

        return self._consumer.name if isinstance(self._consumer, LightningWork) else self._consumer

    @property
    def _path(self) -> Optional[Path]:
        """Path to the file that the payload value gets serialized to."""
        if not self._name:
            return
        return Path("lit://", self._name)

    @abstractmethod
    def save(self, obj: Any, path: str) -> None:
        """Override this method with your own saving logic."""

    @abstractmethod
    def load(self, path: str) -> Any:
        """Override this method with your own loading logic."""

    def _attach_work(self, work: "LightningWork") -> None:
        """Attach a LightningWork to this PayLoad.

        Args:
            work: LightningWork to be attached to this Payload.
        """
        if self._origin is None:
            # Can become an owner only if there is not already one
            self._origin = work.name
        self._consumer = work.name

    def _attach_queues(self, request_queue: BaseQueue, response_queue: BaseQueue) -> None:
        """Attaches the queues for communication with the Storage Orchestrator."""
        self._request_queue = request_queue
        self._response_queue = response_queue

    def _sanitize(self) -> None:
        """Sanitize this Payload so that it can be deep-copied."""
        self._origin = self.origin_name
        self._consumer = self.consumer_name
        self._request_queue = None
        self._response_queue = None

    def exists_remote(self):
        """Check if the payload exists remotely on the attached orgin Work.

        Raises:
            RuntimeError: If the payload is not attached to any Work (origin undefined).
        """
        # Fail early if we need to check the remote but an origin is not defined
        if not self._origin or self._request_queue is None or self._response_queue is None:
            raise RuntimeError(
                f"Trying to check if the payload {self} exists, but the payload is not attached to a LightningWork."
                f" Set it as an attribute to a LightningWork or pass it to the `run()` method."
            )

        # 1. Send message to orchestrator through queue that with a request for a path existence check
        request = _ExistsRequest(source=self.origin_name, name=self._name, path=str(self._path), hash=self.hash)
        self._request_queue.put(request)

        # 2. Wait for the response to come back
        response: _ExistsResponse = self._response_queue.get()  # blocking
        return response.exists

    def get(self) -> Any:
        if _is_flow_context():
            raise RuntimeError("`Payload.get()` can only be called from within the `run()` method of LightningWork.")

        if self._request_queue is None or self._response_queue is None:
            raise RuntimeError(
                f"Trying to get the file {self}, but the payload is not attached to a LightningApp."
                f" Are you trying to get the file from within `__init__`?"
            )
        if self._origin is None:
            raise RuntimeError(
                f"Trying to get the file {self}, but the payload is not attached to a LightningWork. Set it as an"
                f" attribute to a LightningWork or pass it to the `run()` method."
            )

        # 1. Send message to orchestrator through queue with details of the transfer
        # the source is the name of the work that owns the file that we request
        # the destination is determined by the queue, since each work has a dedicated send and recv queue
        request = _GetRequest(source=self.origin_name, name=self._name, path=str(self._path), hash=self.hash)
        self._request_queue.put(request)

        # 2. Wait for the transfer to finish
        response: _GetResponse = self._response_queue.get()  # blocking
        self._validate_get_response(response)

        fs = _filesystem()

        # 3. Wait until the file appears in shared storage
        while not fs.exists(response.path) or fs.info(response.path)["size"] != response.size:
            sleep(REMOTE_STORAGE_WAIT)

        # 4. Copy the file from the shared storage to the destination on the local filesystem
        local_path = self._path
        _logger.debug(f"Attempting to copy {str(response.path)} -> {str(local_path)}")
        fs.get(str(response.path), str(local_path), recursive=False)

        # Ensure the file is properly written
        sleep(0.5)

        self._value = self.load(local_path)
        return self._value

    def _validate_get_response(self, response: "_GetResponse") -> None:
        if response.source != self._origin or response.hash != self.hash:
            raise RuntimeError(
                f"Tried to get the file {self} but received a response for a request it did not send. The response"
                f" contents are: {response}"
            )

        if response.exception is not None:
            raise RuntimeError(
                f"An exception was raised while trying to transfer the contents at {response.path}"
                f" from Work {response.source} to {response.destination}. See the full stacktrace above."
            ) from response.exception

    def to_dict(self) -> dict:
        """Serialize this Path to a dictionary."""
        return dict(
            name=self.name,
            origin_name=self.origin_name,
            consumer_name=self.consumer_name,
            metadata=self._metadata,
        )

    @classmethod
    def from_dict(cls, content: dict) -> "_BasePayload":
        """Instantiate a Payload from a dictionary."""
        payload = cls(None)
        payload._name = content["name"]
        payload._origin = content["origin_name"]
        payload._consumer = content["consumer_name"]
        payload._metadata = content["metadata"]
        return payload

    @staticmethod
    def _handle_exists_request(work: "LightningWork", request: _ExistsRequest) -> _ExistsResponse:
        return _ExistsResponse(
            source=request.source,
            path=request.path,
            name=request.name,
            destination=request.destination,
            hash=request.hash,
            exists=getattr(work, request.name, None) is not None,
        )

    @staticmethod
    def _handle_get_request(work: "LightningWork", request: _GetRequest) -> _GetResponse:
        from lightning.app.storage.copier import _copy_files

        source_path = pathlib.Path(request.path)
        destination_path = _shared_storage_path() / request.hash
        response = _GetResponse(
            source=request.source,
            name=request.name,
            path=str(destination_path),
            hash=request.hash,
            destination=request.destination,
        )

        try:
            payload = getattr(work, request.name)
            payload.save(payload.value, source_path)
            response.size = source_path.stat().st_size
            _copy_files(source_path, destination_path)
            _logger.debug(f"All files copied from {request.path} to {response.path}.")
        except Exception as e:
            response.exception = e
        return response


class Payload(_BasePayload):
    """The Payload object enables to transfer python objects from one work to another in a similar fashion as
    :class:`~lightning.app.storage.path.Path`."""

    def save(self, obj: Any, path: str) -> None:
        with open(path, "wb") as f:
            pickle.dump(obj, f)

    def load(self, path: str) -> Any:
        with open(path, "rb") as f:
            obj = pickle.load(f)
        return obj



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import os
import pdb
import sys

_stdin = [None]
_stdin_lock = multiprocessing.Lock()
try:
    _stdin_fd = sys.stdin.fileno()
except Exception:
    _stdin_fd = None


# Taken from https://github.com/facebookresearch/metaseq/blob/main/metaseq/pdb.py
class MPPdb(pdb.Pdb):
    """A Pdb wrapper that works in a multiprocessing environment."""

    def __init__(self) -> None:
        pdb.Pdb.__init__(self, nosigint=True)

    def _cmdloop(self) -> None:
        stdin_back = sys.stdin
        with _stdin_lock:
            try:
                if _stdin_fd is not None:
                    if not _stdin[0]:
                        _stdin[0] = os.fdopen(_stdin_fd)
                    sys.stdin = _stdin[0]
                self.cmdloop()
            finally:
                sys.stdin = stdin_back


def set_trace() -> None:
    pdb = MPPdb()
    pdb.set_trace(sys._getframe().f_back)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel


class WorkStatus(BaseModel):
    """The ``WorkStatus`` captures the status of a work according to the app."""

    stage: str
    timestamp: float
    reason: Optional[str] = None
    message: Optional[str] = None
    count: int = 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        assert self.timestamp > 0 and self.timestamp < (int(datetime.now().timestamp()) + 10)


class AppStatus(BaseModel):
    """The ``AppStatus`` captures the current status of the app and its components."""

    # ``True`` when the app UI is ready to be viewed
    is_ui_ready: bool

    # The statuses of ``LightningWork`` objects currently associated with this app
    work_statuses: Dict[str, WorkStatus]



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import click

from lightning.app.cli.cmd_apps import _AppManager
from lightning.app.cli.cmd_clusters import AWSClusterManager
from lightning.app.cli.cmd_ssh_keys import _SSHKeyManager


@click.group(name="list")
def get_list() -> None:
    """List Lightning AI self-managed resources (clusters, etc…)"""
    pass


@get_list.command("clusters")
def list_clusters(**kwargs: Any) -> None:
    """List your Lightning AI BYOC compute clusters."""
    cluster_manager = AWSClusterManager()
    cluster_manager.list()


@click.option(
    "--cluster-id",
    "cluster_id",
    type=str,
    required=False,
    default=None,
    help="Filter apps by associated Lightning AI compute cluster",
)
@get_list.command("apps")
def list_apps(cluster_id: str, **kwargs: Any) -> None:
    """List your Lightning AI apps."""
    app_manager = _AppManager()
    app_manager.list(cluster_id=cluster_id)


@get_list.command("ssh-keys")
def list_ssh_keys() -> None:
    """List your Lightning AI ssh-keys."""
    ssh_key_manager = _SSHKeyManager()
    ssh_key_manager.list()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass
from pathlib import Path
from typing import List

__MOUNT_IDENTIFIER__: str = "__mount__"
__MOUNT_PROTOCOLS__: List[str] = ["s3://"]


@dataclass
class Mount:
    """Allows you to mount the contents of an AWS S3 bucket on disk when running an app on the cloud.

    Arguments:
        source: The location which contains the external data which should be mounted in the
            running work. At the moment, only AWS S3 mounts are supported. This must be a full
            `s3` style identifier pointing to a bucket and (optionally) prefix to mount. For
            example: `s3://foo/bar/`.

        mount_path: An absolute directory path in the work where external data source should
            be mounted as a filesystem. This path should not already exist in your codebase.
            If not included, then the root_dir will be set to `/data/<last folder name in the bucket>`
    """

    source: str = ""
    mount_path: str = ""

    def __post_init__(self) -> None:
        for protocol in __MOUNT_PROTOCOLS__:
            if self.source.startswith(protocol):
                protocol = protocol
                break
        else:  # N.B. for-else loop
            raise ValueError(
                f"Unknown protocol for the mount 'source' argument '{self.source}`. The 'source' "
                f"string must start with one of the following prefixes: {__MOUNT_PROTOCOLS__}"
            )

        if protocol == "s3://" and not self.source.endswith("/"):
            raise ValueError(
                "S3 mounts must end in a trailing slash (`/`) to indicate a folder is being mounted. "
                f"Received: '{self.source}'. Mounting a single file is not currently supported."
            )

        if self.mount_path == "":
            self.mount_path = f"/data/{Path(self.source).stem}"

        if not os.path.isabs(self.mount_path):
            raise ValueError(
                f"mount_path argument must be an absolute path to a "
                f"location; received relative path {self.mount_path}"
            )

    @property
    def protocol(self) -> str:
        """The backing storage protocol indicated by this drive source."""
        for protocol in __MOUNT_PROTOCOLS__:
            if self.source.startswith(protocol):
                return protocol
        return ""



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import JSONDecodeError, loads
from typing import Any

from click import ClickException, Context, Group
from lightning_cloud.openapi.rest import ApiException


class _ApiExceptionHandler(Group):
    """Attempts to convert ApiExceptions to ClickExceptions.

    This process clarifies the error for the user by:
    1. Showing the error message from the lightning.ai servers,
       instead of showing the entire HTTP response
    2. Suppressing long tracebacks

    However, if the ApiException cannot be decoded, or is not
    a 4xx error, the original ApiException will be re-raised.
    """

    def invoke(self, ctx: Context) -> Any:
        try:
            return super().invoke(ctx)
        except ApiException as api:
            exception_messages = []
            if 400 <= api.status < 500:
                try:
                    body = loads(api.body)
                except JSONDecodeError:
                    raise api
                exception_messages.append(body["message"])
                exception_messages.extend(body["details"])
            else:
                raise api
            raise ClickException("\n".join(exception_messages))


class MisconfigurationException(Exception):
    """Exception used to inform users of misuse with Lightning."""


class CacheMissException(Exception):
    """Exception used internally as a boundary to non-executed functions."""


class ExitAppException(Exception):
    """Exception used by components to signal that App should exit."""


class LightningComponentException(Exception):
    """Exception used to inform users of misuse with LightningComponent."""


class InvalidPathException(Exception):
    """Exception used to inform users they are accessing an invalid path."""


class LightningFlowException(Exception):
    """Exception used to inform users of misuse with LightningFlow."""


class LightningWorkException(Exception):
    """Exception used to inform users of misuse with LightningWork."""


class LightningPlatformException(Exception):  # pragma: no cover
    """Exception used to inform users of issues related to platform the LightningApp is running on.

    It gets raised by the Lightning Launcher on the platform side when the app is running in the cloud, and is useful
    when framework or user code needs to catch exceptions specific to the platform, e.g., when resources exceed quotas.
    """


class LightningAppStateException(Exception):
    """Exception to inform users of app state errors."""


class LightningSigtermStateException(Exception):
    """Exception to propagate exception in work proxy."""

    def __init__(self, exit_code):
        self.exit_code = exit_code


class LogLinesLimitExceeded(Exception):
    """Exception to inform the user that we've reached the maximum number of log lines."""



from lightning.app.storage.drive import Drive  # noqa: F401
from lightning.app.storage.filesystem import FileSystem  # noqa: F401
from lightning.app.storage.mount import Mount  # noqa: F401
from lightning.app.storage.orchestrator import StorageOrchestrator  # noqa: F401
from lightning.app.storage.path import Path  # noqa: F401
from lightning.app.storage.payload import Payload  # noqa: F401



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for lightning Frontends."""
from __future__ import annotations

import inspect
import os
from typing import Callable

from lightning.app.core.flow import LightningFlow
from lightning.app.utilities.state import AppState


def _reduce_to_flow_scope(state: AppState, flow: str | LightningFlow) -> AppState:
    """Returns a new AppState with the scope reduced to the given flow."""
    flow_name = flow.name if isinstance(flow, LightningFlow) else flow
    flow_name_parts = flow_name.split(".")[1:]  # exclude root
    flow_state = state
    for part in flow_name_parts:
        flow_state = getattr(flow_state, part)
    return flow_state


def _get_flow_state(flow: str) -> AppState:
    """Returns an AppState scoped to the current Flow.

    Returns:
        AppState: An AppState scoped to the current Flow.
    """
    app_state = AppState()
    app_state._request_state()  # pylint: disable=protected-access
    flow_state = _reduce_to_flow_scope(app_state, flow)
    return flow_state


def _get_frontend_environment(flow: str, render_fn_or_file: Callable | str, port: int, host: str) -> os._Environ:
    """Returns an _Environ with the environment variables for serving a Frontend app set.

    Args:
        flow: The name of the flow, for example root.lit_frontend
        render_fn_or_file: A function to render
        port: The port number, for example 54321
        host: The host, for example 'localhost'

    Returns:
        os._Environ: An environment
    """
    env = os.environ.copy()
    env["LIGHTNING_FLOW_NAME"] = flow
    env["LIGHTNING_RENDER_PORT"] = str(port)
    env["LIGHTNING_RENDER_ADDRESS"] = str(host)

    if isinstance(render_fn_or_file, str):
        env["LIGHTNING_RENDER_FILE"] = render_fn_or_file
    else:
        env["LIGHTNING_RENDER_FUNCTION"] = render_fn_or_file.__name__
        env["LIGHTNING_RENDER_MODULE_FILE"] = inspect.getmodule(render_fn_or_file).__file__

    return env



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional

import torch
from torch import Tensor

import lightning.pytorch as pl
from lightning.fabric.plugins import CheckpointIO, ClusterEnvironment
from lightning.fabric.utilities.distributed import _all_gather_ddp_if_available, ReduceOp
from lightning.pytorch.plugins import LayerSync
from lightning.pytorch.plugins.precision import PrecisionPlugin
from lightning.pytorch.strategies.strategy import Strategy


class ParallelStrategy(Strategy, ABC):
    """Plugin for training with multiple processes in parallel."""

    def __init__(
        self,
        accelerator: Optional["pl.accelerators.Accelerator"] = None,
        parallel_devices: Optional[List[torch.device]] = None,
        cluster_environment: Optional[ClusterEnvironment] = None,
        checkpoint_io: Optional[CheckpointIO] = None,
        precision_plugin: Optional[PrecisionPlugin] = None,
    ):
        super().__init__(accelerator=accelerator, checkpoint_io=checkpoint_io, precision_plugin=precision_plugin)
        self.parallel_devices = parallel_devices
        self.cluster_environment: Optional[ClusterEnvironment] = cluster_environment
        self._layer_sync: Optional[LayerSync] = None

    @property
    @abstractmethod
    def root_device(self) -> torch.device:
        """Return the root device."""

    @property
    def global_rank(self) -> int:
        return self.cluster_environment.global_rank() if self.cluster_environment is not None else 0

    @property
    def local_rank(self) -> int:
        return self.cluster_environment.local_rank() if self.cluster_environment is not None else 0

    @property
    def node_rank(self) -> int:
        return self.cluster_environment.node_rank() if self.cluster_environment is not None else 0

    @property
    def world_size(self) -> int:
        return self.cluster_environment.world_size() if self.cluster_environment is not None else 1

    @property
    def is_global_zero(self) -> bool:
        return self.global_rank == 0

    @property
    def parallel_devices(self) -> Optional[List[torch.device]]:
        return self._parallel_devices

    @parallel_devices.setter
    def parallel_devices(self, parallel_devices: Optional[List[torch.device]]) -> None:
        self._parallel_devices = parallel_devices

    @property
    def distributed_sampler_kwargs(self) -> Dict[str, Any]:
        return dict(
            num_replicas=len(self.parallel_devices) if self.parallel_devices is not None else 0,
            rank=self.global_rank,
        )

    def reconciliate_processes(self, trace: str) -> None:
        """Function to re-conciliate processes on failure."""

    def all_gather(self, tensor: Tensor, group: Optional[Any] = None, sync_grads: bool = False) -> Tensor:
        """Perform a all_gather on all processes."""
        return _all_gather_ddp_if_available(tensor, group=group, sync_grads=sync_grads)

    def reduce_boolean_decision(self, decision: bool, all: bool = True) -> bool:
        """Reduces a boolean decision over distributed processes. By default is analagous to ``all`` from the
        standard library, returning ``True`` only if all input decisions evaluate to ``True``. If ``all`` is set to
        ``False``, it behaves like ``any`` instead.

        Args:
            decision: A single input decision.
            all: Whether to logically emulate ``all`` or ``any``. Defaults to True.

        Returns:
            bool: The reduced boolean decision.
        """
        decision = torch.tensor(int(decision), device=self.root_device)
        decision = self.reduce(decision, reduce_op=ReduceOp.SUM)
        decision = bool(decision == self.world_size) if all else bool(decision)
        return decision

    @contextmanager
    def block_backward_sync(self) -> Generator:
        """Blocks ddp sync gradients behaviour on backwards pass.

        This is useful for skipping sync when accumulating gradients, reducing communication overhead
        Returns: context manager with sync behaviour off
        """
        if isinstance(self.model, pl.utilities.types.DistributedDataParallel):
            with self.model.no_sync():
                yield None
        else:
            yield None

    def teardown(self) -> None:
        assert self.cluster_environment is not None
        self.cluster_environment.teardown()
        super().teardown()



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from lightning.pytorch.profilers.advanced import AdvancedProfiler as NewAdvancedProfiler
from lightning.pytorch.profilers.base import PassThroughProfiler as NewPassThroughProfiler
from lightning.pytorch.profilers.profiler import Profiler as NewProfiler
from lightning.pytorch.profilers.pytorch import PyTorchProfiler as NewPyTorchProfiler
from lightning.pytorch.profilers.simple import SimpleProfiler as NewSimpleProfiler
from lightning.pytorch.profilers.xla import XLAProfiler as NewXLAProfiler
from lightning.pytorch.utilities.rank_zero import rank_zero_deprecation


class AdvancedProfiler(NewAdvancedProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.AdvancedProfiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.AdvancedProfiler` class instead."
        )
        super().__init__(*args, **kwargs)


class PassThroughProfiler(NewPassThroughProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.PassThroughProfiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.PassThroughProfiler` class instead."
        )
        super().__init__(*args, **kwargs)


class Profiler(NewProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.Profiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.Profiler` class instead."
        )
        super().__init__(*args, **kwargs)


class PyTorchProfiler(NewPyTorchProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.PyTorchProfiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.PyTorchProfiler` class instead."
        )
        super().__init__(*args, **kwargs)


class SimpleProfiler(NewSimpleProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.SimpleProfiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.SimpleProfiler` class instead."
        )
        super().__init__(*args, **kwargs)


class XLAProfiler(NewXLAProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        rank_zero_deprecation(
            "`lightning.pytorch.profiler.XLAProfiler` is deprecated in v1.9.0 and will be removed in v2.0.0."
            " Use the equivalent `lightning.pytorch.profilers.XLAProfiler` class instead."
        )
        super().__init__(*args, **kwargs)


__all__ = [
    "Profiler",
    "AdvancedProfiler",
    "PassThroughProfiler",
    "PyTorchProfiler",
    "SimpleProfiler",
    "XLAProfiler",
]



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import json
import os
from copy import deepcopy
from time import sleep
from typing import Any, Dict, List, Optional, Tuple, Union

from deepdiff import DeepDiff
from requests import Session
from requests.exceptions import ConnectionError

from lightning.app.core.constants import APP_SERVER_HOST, APP_SERVER_PORT
from lightning.app.storage.drive import _maybe_create_drive
from lightning.app.utilities.app_helpers import AppStatePlugin, BaseStatePlugin, Logger
from lightning.app.utilities.network import _configure_session

logger = Logger(__name__)

# GLOBAL APP STATE
_LAST_STATE = None
_STATE = None


class AppStateType(enum.Enum):
    STREAMLIT = enum.auto()
    DEFAULT = enum.auto()


def headers_for(context: Dict[str, str]) -> Dict[str, str]:
    return {
        "X-Lightning-Session-UUID": context.get("token", ""),
        "X-Lightning-Session-ID": context.get("session_id", ""),
        "X-Lightning-Type": context.get("type", ""),
    }


class AppState:
    _APP_PRIVATE_KEYS: Tuple[str, ...] = (
        "_host",
        "_session_id",
        "_state",
        "_last_state",
        "_url",
        "_port",
        "_request_state",
        "_store_state",
        "_send_state",
        "_my_affiliation",
        "_find_state_under_affiliation",
        "_plugin",
        "_attach_plugin",
        "_authorized",
        "is_authorized",
        "_debug",
        "_session",
    )
    _MY_AFFILIATION: Tuple[str, ...] = ()

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        last_state: Optional[Dict] = None,
        state: Optional[Dict] = None,
        my_affiliation: Tuple[str, ...] = None,
        plugin: Optional[BaseStatePlugin] = None,
    ) -> None:
        """The AppState class enables Frontend users to interact with their application state.

        When the state isn't defined, it would be pulled from the app REST API Server.
        If the state gets modified by the user, the new state would be sent to the API Server.

        Arguments:
            host: Rest API Server current host
            port: Rest API Server current port
            last_state: The state pulled on first access.
            state: The state modified by the user.
            my_affiliation: A tuple describing the affiliation this app state represents. When storing a state dict
                on this AppState, this affiliation will be used to reduce the scope of the given state.
            plugin: A plugin to handle authorization.
        """
        use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ
        self._host = host or APP_SERVER_HOST
        self._port = port or (APP_SERVER_PORT if use_localhost else None)
        self._url = f"{self._host}:{self._port}" if use_localhost else self._host
        self._last_state = last_state
        self._state = state
        self._session_id = "1234"
        self._my_affiliation = my_affiliation if my_affiliation is not None else AppState._MY_AFFILIATION
        self._authorized = None
        self._attach_plugin(plugin)
        self._session = self._configure_session()

    def _attach_plugin(self, plugin: Optional[BaseStatePlugin]) -> None:
        if plugin is not None:
            plugin = plugin
        else:
            plugin = AppStatePlugin()
        self._plugin = plugin

    @staticmethod
    def _find_state_under_affiliation(state, my_affiliation: Tuple[str, ...]) -> Dict[str, Any]:
        """This method is used to extract the subset of the app state associated with the given affiliation.

        For example, if the affiliation is ``("root", "subflow")``, then the returned state will be
        ``state["flows"]["subflow"]``.
        """
        children_state = state
        for name in my_affiliation:
            if name in children_state["flows"]:
                children_state = children_state["flows"][name]
            elif name in children_state["works"]:
                children_state = children_state["works"][name]
            else:
                raise ValueError(f"Failed to extract the state under the affiliation '{my_affiliation}'.")
        return children_state

    def _store_state(self, state: Dict[str, Any]) -> None:
        # Relying on the global variable to ensure the
        # deep_diff is done on the entire state.
        global _LAST_STATE
        global _STATE
        _LAST_STATE = deepcopy(state)
        _STATE = state
        # If the affiliation is passed, the AppState was created in a LightningFlow context.
        # The state should be only the one of this LightningFlow and its children.
        self._last_state = self._find_state_under_affiliation(_LAST_STATE, self._my_affiliation)
        self._state = self._find_state_under_affiliation(_STATE, self._my_affiliation)

    def send_delta(self) -> None:
        app_url = f"{self._url}/api/v1/delta"
        deep_diff = DeepDiff(_LAST_STATE, _STATE, verbose_level=2)
        assert self._plugin is not None
        # TODO: Find how to prevent the infinite loop on refresh without storing the DeepDiff
        if self._plugin.should_update_app(deep_diff):
            data = {"delta": json.loads(deep_diff.to_json())}
            headers = headers_for(self._plugin.get_context())
            try:
                # TODO: Send the delta directly to the REST API.
                response = self._session.post(app_url, json=data, headers=headers)
            except ConnectionError as e:
                raise AttributeError("Failed to connect and send the app state. Is the app running?") from e

            if response and response.status_code != 200:
                raise Exception(f"The response from the server was {response.status_code}. Your inputs were rejected.")

    def _request_state(self) -> None:
        if self._state is not None:
            return
        app_url = f"{self._url}/api/v1/state"
        headers = headers_for(self._plugin.get_context()) if self._plugin else {}

        response_json = {}

        # Sometimes the state URL can return an empty JSON when things are being set-up,
        # so we wait for it to be ready here.
        while response_json == {}:
            sleep(0.5)
            try:
                response = self._session.get(app_url, headers=headers, timeout=1)
            except ConnectionError as e:
                raise AttributeError("Failed to connect and fetch the app state. Is the app running?") from e

            self._authorized = response.status_code
            if self._authorized != 200:
                return

            response_json = response.json()

        logger.debug(f"GET STATE {response} {response_json}")
        self._store_state(response_json)

    def __getattr__(self, name: str) -> Union[Any, "AppState"]:
        if name in self._APP_PRIVATE_KEYS:
            return object.__getattr__(self, name)

        # The state needs to be fetched on access if it doesn't exist.
        self._request_state()

        if name in self._state.get("vars", {}):
            value = self._state["vars"][name]
            if isinstance(value, dict):
                return _maybe_create_drive("root." + ".".join(self._my_affiliation), value)
            return value

        elif name in self._state.get("works", {}):
            return AppState(
                self._host, self._port, last_state=self._last_state["works"][name], state=self._state["works"][name]
            )

        elif name in self._state.get("flows", {}):
            return AppState(
                self._host,
                self._port,
                last_state=self._last_state["flows"][name],
                state=self._state["flows"][name],
            )

        elif name in self._state.get("structures", {}):
            return AppState(
                self._host,
                self._port,
                last_state=self._last_state["structures"][name],
                state=self._state["structures"][name],
            )

        raise AttributeError(
            f"Failed to access '{name}' through `AppState`. The state provides:"
            f" Variables: {list(self._state['vars'].keys())},"
            f" Components: {list(self._state.get('flows', {}).keys()) + list(self._state.get('works', {}).keys())}",
        )

    def __getitem__(self, key: str):
        return self.__getattr__(key)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self._APP_PRIVATE_KEYS:
            object.__setattr__(self, name, value)
            return

        # The state needs to be fetched on access if it doesn't exist.
        self._request_state()

        # TODO: Find a way to aggregate deltas to avoid making
        # request for each attribute change.
        if name in self._state["vars"]:
            self._state["vars"][name] = value
            self.send_delta()

        elif name in self._state["flows"]:
            raise AttributeError("You shouldn't set the flows directly onto the state. Use its attributes instead.")

        elif name in self._state["works"]:
            raise AttributeError("You shouldn't set the works directly onto the state. Use its attributes instead.")

        else:
            raise AttributeError(
                f"Failed to access '{name}' through `AppState`. The state provides:"
                f" Variables: {list(self._state['vars'].keys())},"
                f" Components: {list(self._state['flows'].keys()) + list(self._state['works'].keys())}",
            )

    def __repr__(self) -> str:
        return str(self._state)

    def __bool__(self) -> bool:
        return bool(self._state)

    def __len__(self) -> int:
        # The state needs to be fetched on access if it doesn't exist.
        self._request_state()

        keys = []
        for component in ["flows", "works", "structures"]:
            keys.extend(list(self._state.get(component, {})))
        return len(keys)

    def items(self) -> List[Dict[str, Any]]:
        # The state needs to be fetched on access if it doesn't exist.
        self._request_state()

        items = []
        for component in ["flows", "works"]:
            state = self._state.get(component, {})
            last_state = self._last_state.get(component, {})
            for name, state_value in state.items():
                v = AppState(
                    self._host,
                    self._port,
                    last_state=last_state[name],
                    state=state_value,
                )
                items.append((name, v))

        structures = self._state.get("structures", {})
        last_structures = self._last_state.get("structures", {})
        if structures:
            for component in ["flows", "works"]:
                state = structures.get(component, {})
                last_state = last_structures.get(component, {})
                for name, state_value in state.items():
                    v = AppState(
                        self._host,
                        self._port,
                        last_state=last_state[name],
                        state=state_value,
                    )
                    items.append((name, v))
        return items

    @staticmethod
    def _configure_session() -> Session:
        return _configure_session()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict


def _credential_string_to_basic_auth_params(credential_string: str) -> Dict[str, str]:
    """Returns the name/ID pair for each given Secret name.

    Raises a `ValueError` if any of the given Secret names do not exist.
    """
    if credential_string.count(":") != 1:
        raise ValueError(
            "Credential string must follow the format username:password; "
            + f"the provided one ('{credential_string}') does not."
        )

    username, password = credential_string.split(":")

    if not username:
        raise ValueError("Username cannot be empty.")

    if not password:
        raise ValueError("Password cannot be empty.")

    return {"username": username, "password": password}



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import os
from functools import partial
from pathlib import Path
from shutil import copy2, copystat, Error
from typing import Callable, List, Optional, Set, Tuple, Union

from lightning.app.core.constants import DOT_IGNORE_FILENAME
from lightning.app.utilities.app_helpers import Logger

logger = Logger(__name__)

_IGNORE_FUNCTION = Callable[[Path, List[Path]], List[Path]]


def _copytree(
    src: Union[Path, str],
    dst: Union[Path, str],
    ignore_functions: Optional[List[_IGNORE_FUNCTION]] = None,
    dirs_exist_ok=False,
    dry_run=False,
) -> List[str]:
    """Vendor in from `shutil.copytree` to support ignoring files recursively based on `.lightningignore`, like
    `git` does with `.gitignore`. Also removed a few checks from the original copytree related to symlink checks.
    Differences between original and this function are.

    1. It supports a list of ignore function instead of a single one in the
        original. We can use this for filtering out files based on nested
        .lightningignore files
    2. It supports a dry run. When enabled, this function will not copy anything but just recursively
        find the source files which are not-ignored and return them. It is useful while calculating
        the hash or checking the size of files
    3. This function returns a list of copied files unlike the original which was returning the
        destination directory

    Recursively copy a directory tree and return the destination directory.

    Parameters
    ----------
    src:
        Source directory path to copy from
    dst:
        Destination directory path to copy to
    ignore_functions:
        List of functions that will be used to filter out files
        and directories. This isn't required to be passed when calling from outside but will be
        autopopulated by the recursive calls in this function itself (Original copytree doesn't have this argument)
    dirs_exist_ok:
        If true, the destination directory will be created if it doesn't exist.
    dry_run:
        If true, this function will not copy anything (this is not present in the original copytree)


    If exception(s) occur, an Error is raised with a list of reasons.
    """
    files_copied = []

    if ignore_functions is None:
        ignore_functions = []

    _ignore_filename_spell_check(src)
    src = Path(src)
    dst = Path(dst)
    ignore_filepath = src / DOT_IGNORE_FILENAME
    if ignore_filepath.is_file():
        patterns = _read_lightningignore(ignore_filepath)
        ignore_fn = partial(_filter_ignored, src, patterns)
        # creating new list so we won't modify the original
        ignore_functions = [*ignore_functions, ignore_fn]

    if not dry_run:
        os.makedirs(dst, exist_ok=dirs_exist_ok)

    errors = []

    entries = [dr for dr in src.iterdir()]
    for fn in ignore_functions:
        # ignore function return only the entries that are not ignored
        entries = fn(src, entries)

    for srcentry in entries:
        dstpath = dst / srcentry.name
        try:
            if srcentry.is_dir():
                _files = _copytree(
                    src=srcentry,
                    dst=dstpath,
                    ignore_functions=ignore_functions,
                    dirs_exist_ok=dirs_exist_ok,
                    dry_run=dry_run,
                )
                files_copied.extend(_files)
            else:
                files_copied.append(str(srcentry))
                if not dry_run:
                    # Will raise a SpecialFileError for unsupported file types
                    copy2(srcentry, dstpath)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcentry, dstpath, str(why)))
    try:
        if not dry_run:
            copystat(src, dst)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, "winerror", None) is None:
            errors.append((src, dst, str(why)))
    if errors:
        raise Error(errors)
    return files_copied


def _filter_ignored(src: Path, patterns: Set[str], current_dir: Path, entries: List[Path]) -> List[Path]:
    relative_dir = current_dir.relative_to(src)
    names = [str(relative_dir / entry.name) for entry in entries]
    ignored_names = set()
    for pattern in patterns:
        ignored_names.update(fnmatch.filter(names, pattern))
    return [entry for entry in entries if str(relative_dir / entry.name) not in ignored_names]


def _parse_lightningignore(lines: Tuple[str]) -> Set[str]:
    """Creates a set that removes empty lines and comments."""
    lines = [ln.strip() for ln in lines]
    # removes first `/` character for posix and `\\` for windows
    lines = [ln.lstrip("/").lstrip("\\") for ln in lines if ln != "" and not ln.startswith("#")]
    # convert to path and converting back to string to sanitize the pattern
    return {str(Path(ln)) for ln in lines}


def _read_lightningignore(path: Path) -> Set[str]:
    """Reads ignore file and filter and empty lines. This will also remove patterns that start with a `/`. That's
    done to allow `glob` to simulate the behavior done by `git` where it interprets that as a root path.

    Parameters
    ----------
    path: Path
        Path to .lightningignore file or equivalent.

    Returns
    -------
    Set[str]
        Set of unique lines.
    """
    raw_lines = path.open().readlines()
    return _parse_lightningignore(raw_lines)


def _ignore_filename_spell_check(src: Path):
    possible_spelling_mistakes = [
        ".gridignore",
        ".lightingignore",
        ".lightinginore",
        ".lightninginore",
        ".lightninignore",
        ".lightinignore",
    ]
    possible_spelling_mistakes.extend([p.lstrip(".") for p in possible_spelling_mistakes])
    for path in src.iterdir():
        if path.is_file() and path.name in possible_spelling_mistakes:
            logger.warn(
                f"Lightning uses `{DOT_IGNORE_FILENAME}` as the ignore file but found {path.name} at "
                f"{path.parent} instead. If this was a mistake, please rename the file."
            )



import random

from lightning_cloud.openapi import V1ClusterType, V1ProjectClusterBinding
from lightning_cloud.openapi.rest import ApiException

from lightning.app.utilities.network import LightningClient


def _ensure_cluster_project_binding(client: LightningClient, project_id: str, cluster_id: str) -> None:
    cluster_bindings = client.projects_service_list_project_cluster_bindings(project_id=project_id)

    for cluster_binding in cluster_bindings.clusters:
        if cluster_binding.cluster_id != cluster_id:
            continue
        if cluster_binding.project_id == project_id:
            return

    client.projects_service_create_project_cluster_binding(
        project_id=project_id,
        body=V1ProjectClusterBinding(cluster_id=cluster_id, project_id=project_id),
    )


def _get_default_cluster(client: LightningClient, project_id: str) -> str:
    """This utility implements a minimal version of the cluster selection logic used in the cloud.

    TODO: This should be requested directly from the platform.
    """
    cluster_bindings = client.projects_service_list_project_cluster_bindings(project_id=project_id).clusters

    if not cluster_bindings:
        raise ValueError(f"No clusters are bound to the project {project_id}.")

    if len(cluster_bindings) == 1:
        return cluster_bindings[0].cluster_id

    clusters = []
    for cluster_binding in cluster_bindings:
        try:
            clusters.append(client.cluster_service_get_cluster(cluster_binding.cluster_id))
        except ApiException:
            # If we failed to get the cluster, ignore it
            continue

    # Filter global clusters
    clusters = [cluster for cluster in clusters if cluster.spec.cluster_type == V1ClusterType.GLOBAL]

    if len(clusters) == 0:
        raise RuntimeError(f"No clusters found on `{client.api_client.configuration.host}`.")

    return random.choice(clusters).id



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import warnings
from typing import Dict, List, Union

import lightning.app
from lightning.app.frontend.frontend import Frontend
from lightning.app.utilities.app_helpers import _MagicMockJsonSerializable, is_overridden
from lightning.app.utilities.cloud import is_running_in_cloud


def _add_comment_to_literal_code(method, contains, comment):
    """Inspects a method's code and adds a message to it.

    This is a nice to have, so if it fails for some reason, it shouldn't affect the program.
    """
    try:
        lines = inspect.getsource(method)
        lines = lines.split("\n")
        idx_list = [i for i, x in enumerate(lines) if contains in x]
        for i in idx_list:
            line = lines[i]
            line += comment
            lines[i] = line

        return "\n".join(lines)

    except Exception as e:  # noqa
        return ""


def _collect_layout(app: "lightning.app.LightningApp", flow: "lightning.app.LightningFlow") -> Union[Dict, List[Dict]]:
    """Process the layout returned by the ``configure_layout()`` method in each flow."""
    layout = flow.configure_layout()

    if isinstance(layout, Frontend):
        frontend = layout
        frontend.flow = flow
        app.frontends.setdefault(flow.name, frontend)

        # When running locally, the target will get overwritten by the dispatcher when launching the frontend servers
        # When running in the cloud, the frontend code will construct the URL based on the flow name
        return flow._layout
    elif isinstance(layout, _MagicMockJsonSerializable):
        # The import was mocked, we set a dummy `Frontend` so that `is_headless` knows there is a UI
        app.frontends.setdefault(flow.name, "mock")
        return flow._layout
    elif isinstance(layout, dict):
        layout = _collect_content_layout([layout], app, flow)
    elif isinstance(layout, (list, tuple)) and all(isinstance(item, dict) for item in layout):
        layout = _collect_content_layout(layout, app, flow)
    else:
        lines = _add_comment_to_literal_code(flow.configure_layout, contains="return", comment="  <------- this guy")
        m = f"""
        The return value of configure_layout() in `{flow.__class__.__name__}`  is an unsupported layout format:
        \n{lines}

        Return either an object of type {Frontend} (e.g., StreamlitFrontend, StaticWebFrontend):
            def configure_layout(self):
                return la.frontend.Frontend(...)

        OR a single dict:
            def configure_layout(self):
                tab1 = {{'name': 'tab name', 'content': self.a_component}}
                return tab1

        OR a list of dicts:
            def configure_layout(self):
                tab1 = {{'name': 'tab name 1', 'content': self.component_a}}
                tab2 = {{'name': 'tab name 2', 'content': self.component_b}}
                return [tab1, tab2]

        (see the docs for `LightningFlow.configure_layout`).
        """
        raise TypeError(m)

    return layout


def _collect_content_layout(
    layout: List[Dict], app: "lightning.app.LightningApp", flow: "lightning.app.LightningFlow"
) -> Union[List[Dict], Dict]:
    """Process the layout returned by the ``configure_layout()`` method if the returned format represents an
    aggregation of child layouts."""
    for entry in layout:
        if "content" not in entry:
            raise ValueError(
                f"A dictionary returned by `{flow.__class__.__name__}.configure_layout()` is missing a key 'content'."
                f" For the value, choose either a reference to a child flow or a URla."
            )
        if isinstance(entry["content"], str):  # assume this is a URL
            url = entry["content"]
            if url.startswith("/"):
                # The URL isn't fully defined yet. Looks something like ``self.work.url + /something``.
                entry["target"] = ""
            else:
                entry["target"] = url
            if url.startswith("http://") and is_running_in_cloud():
                warnings.warn(
                    f"You configured an http link {url[:32]}... but it won't be accessible in the cloud."
                    f" Consider replacing 'http' with 'https' in the link above."
                )

        elif isinstance(entry["content"], lightning.app.LightningFlow):
            entry["content"] = entry["content"].name

        elif isinstance(entry["content"], lightning.app.LightningWork):
            work = entry["content"]
            work_layout = _collect_work_layout(work)

            if work_layout is None:
                entry["content"] = ""
            elif isinstance(work_layout, str):
                entry["content"] = work_layout
                entry["target"] = work_layout
            elif isinstance(work_layout, (Frontend, _MagicMockJsonSerializable)):
                if len(layout) > 1:
                    lines = _add_comment_to_literal_code(
                        flow.configure_layout, contains="return", comment="  <------- this guy"
                    )
                    m = f"""
                    The return value of configure_layout() in `{flow.__class__.__name__}`  is an
                    unsupported format:
                    \n{lines}

                    The tab containing a `{work.__class__.__name__}` must be the only tab in the
                    layout of this flow.

                    (see the docs for `LightningWork.configure_layout`).
                    """
                    raise TypeError(m)

                if isinstance(work_layout, Frontend):
                    # If the work returned a frontend, treat it as belonging to the flow.
                    # NOTE: This could evolve in the future to run the Frontend directly in the work machine.
                    frontend = work_layout
                    frontend.flow = flow
                elif isinstance(work_layout, _MagicMockJsonSerializable):
                    # The import was mocked, we set a dummy `Frontend` so that `is_headless` knows there is a UI.
                    frontend = "mock"

                app.frontends.setdefault(flow.name, frontend)
                return flow._layout

        elif isinstance(entry["content"], _MagicMockJsonSerializable):
            # The import was mocked, we just record dummy content so that `is_headless` knows there is a UI
            entry["content"] = "mock"
            entry["target"] = "mock"
        else:
            m = f"""
            A dictionary returned by `{flow.__class__.__name__}.configure_layout()` contains an unsupported entry.

            {{'content': {repr(entry['content'])}}}

            Set the `content` key to a child flow or a URL, for example:

            class {flow.__class__.__name__}(LightningFlow):
                def configure_layout(self):
                    return {{'content': childFlow OR childWork OR 'http://some/url'}}
            """
            raise ValueError(m)
    return layout


def _collect_work_layout(work: "lightning.app.LightningWork") -> Union[None, str, Frontend, _MagicMockJsonSerializable]:
    """Check if ``configure_layout`` is overridden on the given work and return the work layout (either a string, a
    ``Frontend`` object, or an instance of a mocked import).

    Args:
        work: The work to collect the layout for.

    Raises:
        TypeError: If the value returned by ``configure_layout`` is not of a supported format.
    """
    if is_overridden("configure_layout", work):
        work_layout = work.configure_layout()
    else:
        work_layout = work.url

    if work_layout is None:
        return None
    elif isinstance(work_layout, str):
        url = work_layout
        # The URL isn't fully defined yet. Looks something like ``self.work.url + /something``.
        if url and not url.startswith("/"):
            return url
        return ""
    elif isinstance(work_layout, (Frontend, _MagicMockJsonSerializable)):
        return work_layout
    else:
        m = f"""
        The value returned by `{work.__class__.__name__}.configure_layout()` is of an unsupported type.

        {repr(work_layout)}

        Return a `Frontend` or a URL string, for example:

        class {work.__class__.__name__}(LightningWork):
            def configure_layout(self):
                return MyFrontend() OR 'http://some/url'
        """
        raise TypeError(m)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as t

from lightning.app.utilities.app_helpers import _LightningAppRef, _set_child_name

T = t.TypeVar("T")

if t.TYPE_CHECKING:
    from lightning.app.utilities.types import Component


def _prepare_name(component: "Component") -> str:
    return str(component.name.split(".")[-1])


# TODO: add support and tests for list operations (concatenation, deletion, insertion, etc.)
class List(t.List[T]):
    def __init__(self, *items: T):
        """The List Object is used to represents list collection of
        :class:`~lightning.app.core.work.LightningWork`
        or :class:`~lightning.app.core.flow.LightningFlow`.

        Example:

            >>> from lightning.app import LightningFlow, LightningWork
            >>> from lightning.app.structures import List
            >>> class CounterWork(LightningWork):
            ...     def __init__(self):
            ...         super().__init__()
            ...         self.counter = 0
            ...     def run(self):
            ...         self.counter += 1
            ...
            >>> class RootFlow(LightningFlow):
            ...     def __init__(self):
            ...         super().__init__()
            ...         self.list = List(*[CounterWork(), CounterWork()])
            ...     def run(self):
            ...         for work in self.list:
            ...             work.run()
            ...
            >>> flow = RootFlow()
            >>> flow.run()
            >>> assert flow.list[0].counter == 1

        Arguments:
            items: A sequence of LightningWork or LightningFlow.
        """
        super().__init__()
        from lightning.app.runners.backends import Backend

        self._name: t.Optional[str] = ""
        self._last_index = 0
        self._backend: t.Optional[Backend] = None
        for item in items:
            self.append(item)

    def append(self, v):
        from lightning.app import LightningFlow, LightningWork

        _set_child_name(self, v, str(self._last_index))
        if self._backend:
            if isinstance(v, LightningFlow):
                LightningFlow._attach_backend(v, self._backend)
            elif isinstance(v, LightningWork):
                self._backend._wrap_run_method(_LightningAppRef().get_current(), v)
        v._name = f"{self.name}.{self._last_index}"
        self._last_index += 1
        super().append(v)

    @property
    def name(self):
        """Returns the name of this List object."""
        return self._name or "root"

    @property
    def works(self):
        from lightning.app import LightningFlow, LightningWork

        works = [item for item in self if isinstance(item, LightningWork)]
        for flow in [item for item in self if isinstance(item, LightningFlow)]:
            for child_work in flow.works(recurse=False):
                works.append(child_work)
        return works

    @property
    def flows(self):
        from lightning.app import LightningFlow
        from lightning.app.structures import Dict, List

        flows = {}
        for item in self:
            if isinstance(item, LightningFlow):
                flows[item.name] = item
                for child_flow in item.flows.values():
                    flows[child_flow.name] = child_flow
            if isinstance(item, (Dict, List)):
                for child_flow in item.flows.values():
                    flows[child_flow.name] = child_flow
        return flows

    @property
    def state(self):
        """Returns the state of its flows and works."""
        from lightning.app import LightningFlow, LightningWork

        works = [item for item in self if isinstance(item, LightningWork)]
        children = [item for item in self if isinstance(item, LightningFlow)]
        return {
            "works": {_prepare_name(w): w.state for w in works},
            "flows": {_prepare_name(flow): flow.state for flow in children},
        }

    @property
    def state_vars(self):
        from lightning.app import LightningFlow, LightningWork

        works = [item for item in self if isinstance(item, LightningWork)]
        children = [item for item in self if isinstance(item, LightningFlow)]
        return {
            "works": {_prepare_name(w): w.state_vars for w in works},
            "flows": {_prepare_name(flow): flow.state_vars for flow in children},
        }

    @property
    def state_with_changes(self):
        from lightning.app import LightningFlow, LightningWork

        works = [item for item in self if isinstance(item, LightningWork)]
        children = [item for item in self if isinstance(item, LightningFlow)]
        return {
            "works": {str(_prepare_name(w)): w.state_with_changes for w in works},
            "flows": {_prepare_name(flow): flow.state_with_changes for flow in children},
        }

    def set_state(self, state):
        """Method to set the state of the list and its children."""
        from lightning.app import LightningFlow, LightningWork

        works = [item for item in self if isinstance(item, LightningWork)]
        children = [item for item in self if isinstance(item, LightningFlow)]

        current_state_keys = {_prepare_name(w) for w in self}
        state_keys = set(list(state["works"].keys()) + list(state["flows"].keys()))

        if current_state_keys != state_keys:
            key_diff = (current_state_keys - state_keys) | (state_keys - current_state_keys)
            raise Exception(
                f"The provided state doesn't match the `List` {self.name}. Found `{key_diff}` un-matching keys"
            )

        for work_key, work_state in state["works"].items():
            for work in works:
                if _prepare_name(work) == work_key:
                    work.set_state(work_state)
        for child_key, child_state in state["flows"].items():
            for child in children:
                if _prepare_name(child) == child_key:
                    child.set_state(child_state)

    def __len__(self):
        """Returns the number of elements within this List."""
        return len([v for v in self])



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
import json
import os
import shutil
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager, nullcontext
from multiprocessing import Process
from subprocess import Popen
from time import sleep
from typing import Any, Callable, Dict, Generator, List, Optional, Type

import requests
from lightning_cloud.openapi import V1LightningappInstanceState
from lightning_cloud.openapi.rest import ApiException
from requests import Session
from rich import print
from rich.color import ANSI_COLOR_NAMES

from lightning.app import LightningApp, LightningFlow
from lightning.app.cli.lightning_cli import run_app
from lightning.app.core.constants import LIGHTNING_CLOUD_PROJECT_ID
from lightning.app.runners.multiprocess import MultiProcessRuntime
from lightning.app.testing.config import _Config
from lightning.app.utilities.app_logs import _app_logs_reader
from lightning.app.utilities.cloud import _get_project
from lightning.app.utilities.enum import CacheCallsKeys
from lightning.app.utilities.imports import _is_playwright_available, requires
from lightning.app.utilities.log import get_logfile
from lightning.app.utilities.logs_socket_api import _LightningLogsSocketAPI
from lightning.app.utilities.network import _configure_session, LightningClient
from lightning.app.utilities.packaging.lightning_utils import get_dist_path_if_editable_install
from lightning.app.utilities.proxies import ProxyWorkRun

if _is_playwright_available():
    import playwright
    from playwright.sync_api import HttpCredentials, sync_playwright


def _on_error_callback(ws_app, *_):
    ws_app.close()


def _print_logs(app_id: str):
    client = LightningClient()
    project = _get_project(client)

    works = client.lightningwork_service_list_lightningwork(
        project_id=project.project_id,
        app_id=app_id,
    ).lightningworks
    component_names = ["flow"] + [w.name for w in works]

    rich_colors = list(ANSI_COLOR_NAMES)
    colors = {c: rich_colors[i + 1] for i, c in enumerate(component_names)}

    max_length = max(len(c.replace("root.", "")) for c in component_names)
    identifiers = []

    print("################### PRINTING LOGS ###################")

    logs_api_client = _LightningLogsSocketAPI(client.api_client)

    while True:
        gen = _app_logs_reader(
            logs_api_client=logs_api_client,
            project_id=project.project_id,
            app_id=app_id,
            component_names=component_names,
            follow=False,
            on_error_callback=_on_error_callback,
        )
        for log_event in gen:
            message = log_event.message
            identifier = f"{log_event.timestamp}{log_event.message}"
            if identifier not in identifiers:
                date = log_event.timestamp.strftime("%m/%d/%Y %H:%M:%S")
                identifiers.append(identifier)
                color = colors[log_event.component_name]
                padding = (max_length - len(log_event.component_name)) * " "
                print(f"[{color}]{log_event.component_name}{padding}[/{color}] {date} {message}")


class LightningTestApp(LightningApp):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.counter = 0

    @staticmethod
    def _configure_session() -> Session:
        return _configure_session()

    def make_request(self, fn, *args, **kwargs):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._make_request(fn, *args, **kwargs))

    async def _make_request(self, fn: Callable, *args, **kwargs):
        from lightning.app.utilities.state import AppState

        state = AppState()
        state._request_state()
        fn(state, *args, **kwargs)
        state.send_delta()

    def on_before_run_once(self):
        pass

    def on_after_run_once(self):
        self.counter += 1

    def run_once(self):
        before_done = self.on_before_run_once()
        if before_done is not None:
            return before_done
        done = super().run_once()
        after_done = self.on_after_run_once()
        if after_done is not None:
            return after_done
        return done

    def kill_work(self, work_name: str, sleep_time: int = 1):
        """Use this method to kill a specific work by its name."""
        self.processes[work_name].kill()

    def restart_work(self, work_name: str):
        """Use this method to restart a specific work by its name."""
        self.processes[work_name].restart()


@requires("click")
def application_testing(
    lightning_app_cls: Type[LightningTestApp] = LightningTestApp, command_line: List[str] = []
) -> Any:
    from unittest import mock

    from click.testing import CliRunner

    patch1 = mock.patch("lightning.app.LightningApp", lightning_app_cls)
    # we need to patch both only with the mirror package
    patch2 = (
        mock.patch("lightning.LightningApp", lightning_app_cls) if "lightning.app" in sys.modules else nullcontext()
    )
    with patch1, patch2:
        original = sys.argv
        sys.argv = command_line
        runner = CliRunner()
        result = runner.invoke(run_app, command_line, catch_exceptions=False)
        sys.argv = original
        return result


class _SingleWorkFlow(LightningFlow):
    def __init__(self, work, args, kwargs):
        super().__init__()
        self.work = work
        self.args = args
        self.kwargs = kwargs

    def run(self):
        if self.work.has_succeeded or self.work.has_failed:
            self.stop()
        self.work.run(*self.args, **self.kwargs)


def run_work_isolated(work, *args, start_server: bool = False, **kwargs):
    """This function is used to run a work a single time with multiprocessing runtime."""
    MultiProcessRuntime(
        LightningApp(_SingleWorkFlow(work, args, kwargs), log_level="debug"),
        start_server=start_server,
    ).dispatch()
    # pop the stopped status.
    call_hash = work._calls[CacheCallsKeys.LATEST_CALL_HASH]

    if call_hash in work._calls:
        work._calls[call_hash]["statuses"].pop(-1)

    if isinstance(work.run, ProxyWorkRun):
        work.run = work.run.work_run


def _browser_context_args(browser_context_args: Dict) -> Dict:
    return {
        **browser_context_args,
        "viewport": {
            "width": 1920,
            "height": 1080,
        },
        "ignore_https_errors": True,
    }


@contextmanager
def _run_cli(args) -> Generator:
    """This utility is used to automate end-to-end testing of the Lightning AI CLI."""
    cmd = [
        sys.executable,
        "-m",
        "lightning",
    ] + args

    with tempfile.TemporaryDirectory() as tmpdir:
        env_copy = os.environ.copy()
        process = Popen(
            cmd,
            cwd=tmpdir,
            env=env_copy,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        process.wait()

    yield process.stdout.read().decode("UTF-8"), process.stderr.read().decode("UTF-8")


def _fetch_app_by_name(client, project_id, name):
    lit_apps = [
        app
        for app in client.lightningapp_instance_service_list_lightningapp_instances(project_id=project_id).lightningapps
        if app.name == name or getattr(app, "display_name", None) == name
    ]
    if not len(lit_apps) == 1:
        raise ValueError(f"Expected to find just one app, found {len(lit_apps)}")
    app = lit_apps[0]
    return app


@requires("playwright")
@contextmanager
def run_app_in_cloud(
    app_folder: str, app_name: str = "app.py", extra_args: List[str] = [], debug: bool = True
) -> Generator:
    """This utility is used to automate testing e2e application with lightning.ai."""
    # 1. Validate the provide app_folder is correct.
    if not os.path.exists(os.path.join(app_folder, "app.py")):
        raise Exception("The app folder should contain an app.py file.")
    if app_folder.endswith("/"):
        app_folder = app_folder[:-1]

    # 2. Create the right application name.
    basename = app_folder.split("/")[-1]
    PR_NUMBER = os.getenv("PR_NUMBER", None)

    is_editable_mode = get_dist_path_if_editable_install("lightning")
    if not is_editable_mode and PR_NUMBER is not None:
        raise Exception("Lightning requires to be installed in editable mode in the CI.")

    TEST_APP_NAME = os.getenv("TEST_APP_NAME", basename)
    os.environ["TEST_APP_NAME"] = TEST_APP_NAME

    if PR_NUMBER:
        name = f"test-{PR_NUMBER}-{TEST_APP_NAME}-" + str(int(time.time()))
    else:
        name = f"test-{TEST_APP_NAME}-" + str(int(time.time()))

    os.environ["LIGHTNING_APP_NAME"] = name

    url = _Config.url
    if url.endswith("/"):
        url = url[:-1]
    payload = {"apiKey": _Config.api_key, "username": _Config.username}
    url_login = url + "/v1/auth/login"
    res = requests.post(url_login, data=json.dumps(payload))
    if "token" not in res.json():
        raise RuntimeError(
            f"You haven't properly setup your environment variables with {url_login} and data: \n{payload}"
        )

    token = res.json()["token"]

    # 3. Disconnect from the App if any.
    Popen("lightning logout", shell=True).wait()

    # 4. Launch the application in the cloud from the Lightning CLI.
    with tempfile.TemporaryDirectory() as tmpdir:
        env_copy = os.environ.copy()
        env_copy["PACKAGE_LIGHTNING"] = "1"
        env_copy["LIGHTING_TESTING"] = "1"
        if debug:
            print("Debug mode is enabled")
            env_copy["LIGHTNING_DEBUG"] = "1"
        shutil.copytree(app_folder, tmpdir, dirs_exist_ok=True)
        # TODO - add -no-cache to the command line.
        stdout_path = get_logfile(f"run_app_in_cloud_{name}")
        with open(stdout_path, "w") as stdout:
            cmd = [
                sys.executable,
                "-m",
                "lightning",
                "run",
                "app",
                app_name,
                "--cloud",
                "--name",
                name,
                "--open-ui",
                "false",
            ]
            process = Popen((cmd + extra_args), cwd=tmpdir, env=env_copy, stdout=stdout, stderr=sys.stderr)
            process.wait()

        if is_editable_mode:
            # Added to ensure the current code is properly uploaded.
            # Otherwise, it could result in un-tested PRs.
            pkg_found = False
            with open(stdout_path) as fo:
                for line in fo.readlines():
                    if "Packaged Lightning with your application" in line:
                        pkg_found = True
                    print(line)  # TODO: use logging
            assert pkg_found
        os.remove(stdout_path)

    # 5. Print your application name
    print(f"The Lightning App Name is: [bold magenta]{name}[/bold magenta]")

    # 6. Create chromium browser, auth to lightning.app.ai and yield the admin and view pages.
    with sync_playwright() as p:
        browser = p.chromium.launch(headless=bool(int(os.getenv("HEADLESS", "0"))))
        context = browser.new_context(
            # Eventually this will need to be deleted
            http_credentials=HttpCredentials(
                {"username": os.getenv("LAI_USER", "").strip(), "password": os.getenv("LAI_PASS", "")}
            ),
            record_video_dir=os.path.join(_Config.video_location, TEST_APP_NAME),
            record_har_path=_Config.har_location,
        )
        admin_page = context.new_page()
        print(f"The Lightning App Token is: {token}")
        print(f"The Lightning App user key is: {_Config.key}")
        print(f"The Lightning App user id is: {_Config.id}")
        admin_page.goto(_Config.url)
        admin_page.evaluate(
            """data => {
            window.localStorage.setItem('gridUserId', data[0]);
            window.localStorage.setItem('gridUserKey', data[1]);
            window.localStorage.setItem('gridUserToken', data[2]);
        }
        """,
            [_Config.id, _Config.key, token],
        )
        if LIGHTNING_CLOUD_PROJECT_ID:
            admin_page.evaluate(
                """data => {
                window.localStorage.setItem('gridDefaultProjectIdOverride', JSON.stringify(data[0]));
            }
            """,
                [LIGHTNING_CLOUD_PROJECT_ID],
            )
        admin_page.goto(f"{_Config.url}/{_Config.username}/apps", timeout=60 * 1000)

        # Closing the Complete your profile dialog
        try:
            dialog = admin_page.locator("text=Complete your profile")
            dialog.wait_for(timeout=10 * 1000, state="visible")
            print("'Complete your profile' dialog visible, closing it.")
            admin_page.locator('input[name="firstName"]').fill("first")
            admin_page.locator('input[name="lastName"]').fill("last")
            admin_page.locator('input[name="email"]').fill("e2e.test.admin@lightning.ai")
            admin_page.locator('input[name="organization"]').fill("Lightning AI")
            button = admin_page.locator('button:has-text("Confirm")')
            button.wait_for(timeout=3 * 1000)
            button.click()
        except playwright._impl._api_types.TimeoutError:
            print("'Complete your profile' dialog not visible, skipping.")

        # Closing the Create Project dialog.
        try:
            project_dialog = admin_page.locator("text=Create a project")
            project_dialog.wait_for(timeout=10 * 1000, state="visible")
            print("'Create Project' dialog visible, closing it.")
            project_name_input = admin_page.locator('input[type="text"]')
            project_name_input.fill("Default Project")
            button = admin_page.locator('button:has-text("Continue")')
            button.wait_for(timeout=3 * 1000)
            button.click()
        except playwright._impl._api_types.TimeoutError:
            print("'Create Project' dialog not visible, skipping.")

        admin_page.locator(f'[data-cy="{name}"]').click()

        app_url = admin_page.url
        admin_page.goto(app_url + "/logs")

        client = LightningClient()
        project_id = _get_project(client).project_id

        app = _fetch_app_by_name(client, project_id, name)
        app_id = app.id
        print(f"The Lightning App ID is: {app.id}")  # useful for Grafana

        if debug:
            process = Process(target=_print_logs, kwargs={"app_id": app_id})
            process.start()

        view_page = context.new_page()
        i = 1
        while True:
            app = _fetch_app_by_name(client, project_id, name)
            msg = f"Still in phase {app.status.phase}"

            # wait until the app is running and openapi.json is ready
            if app.status.phase == V1LightningappInstanceState.RUNNING:
                view_page.goto(f"{app.status.url}/view")
                status_code = requests.get(f"{app.status.url}/openapi.json").status_code
                if status_code == 200:
                    print("App is running, continuing with testing...")
                    break
                msg = f"Received status code {status_code} at {app.status.url!r}"
            elif app.status.phase not in (V1LightningappInstanceState.PENDING, V1LightningappInstanceState.NOT_STARTED):
                # there's a race condition if the app goes from pending to running to something else before we evaluate
                # the condition above. avoid it by checking stopped explicitly
                print(f"App finished with phase {app.status.phase}, finished testing...")
                break
            if debug and i % 30 == 0:
                print(f"{msg}, continuing infinite loop...")
            i += 1
            sleep(1)

        logs_api_client = _LightningLogsSocketAPI(client.api_client)

        def fetch_logs(component_names: Optional[List[str]] = None) -> Generator:
            """This methods creates websockets connection in threads and returns the logs to the main thread."""
            if not component_names:
                works = client.lightningwork_service_list_lightningwork(
                    project_id=project_id,
                    app_id=app_id,
                ).lightningworks

                component_names = ["flow"] + [w.name for w in works]
            else:

                def add_prefix(c: str) -> str:
                    if c == "flow":
                        return c
                    if not c.startswith("root."):
                        return "root." + c
                    return c

                component_names = [add_prefix(c) for c in component_names]

            gen = _app_logs_reader(
                logs_api_client=logs_api_client,
                project_id=project_id,
                app_id=app_id,
                component_names=component_names,
                follow=False,
                on_error_callback=_on_error_callback,
            )
            for log_event in gen:
                yield log_event.message

        try:
            yield admin_page, view_page, fetch_logs, name
        except KeyboardInterrupt:
            pass
        finally:
            if debug:
                process.kill()

            context.close()
            browser.close()
            Popen("lightning disconnect", shell=True).wait()


def wait_for(page, callback: Callable, *args, **kwargs) -> Any:
    import playwright

    while True:
        try:
            res = callback(*args, **kwargs)
            if res:
                return res
        except (playwright._impl._api_types.Error, playwright._impl._api_types.TimeoutError) as err:
            print(err)
            try:
                sleep(7)
                page.reload()
            except (playwright._impl._api_types.Error, playwright._impl._api_types.TimeoutError) as err:
                print(err)
                pass
            sleep(3)


def _delete_lightning_app(client, project_id, app_id, app_name):
    print(f"Deleting {app_name} id: {app_id}")
    try:
        res = client.lightningapp_instance_service_delete_lightningapp_instance(
            project_id=project_id,
            id=app_id,
        )
        assert res == {}
    except ApiException as ex:
        print(f"Failed to delete {app_name}. Exception {ex}")


def delete_cloud_lightning_apps():
    """Cleanup cloud apps that start with the name test-{PR_NUMBER}-{TEST_APP_NAME}.

    PR_NUMBER and TEST_APP_NAME are environment variables.
    """

    client = LightningClient()

    try:
        pr_number = int(os.getenv("PR_NUMBER", None))
    except (TypeError, ValueError):
        # Failed when the PR is running master or 'PR_NUMBER' isn't defined.
        pr_number = ""

    app_name = os.getenv("TEST_APP_NAME", "")

    print(f"deleting apps for pr_number: {pr_number}, app_name: {app_name}")
    project_id = _get_project(client).project_id
    list_apps = client.lightningapp_instance_service_list_lightningapp_instances(project_id=project_id)

    for lit_app in list_apps.lightningapps:
        if pr_number and app_name and not lit_app.name.startswith(f"test-{pr_number}-{app_name}-"):
            continue
        _delete_lightning_app(client, project_id=project_id, app_id=lit_app.id, app_name=lit_app.name)

    print("deleting apps that were created more than 1 hour ago.")

    for lit_app in list_apps.lightningapps:

        if lit_app.created_at < datetime.datetime.now(lit_app.created_at.tzinfo) - datetime.timedelta(hours=1):
            _delete_lightning_app(client, project_id=project_id, app_id=lit_app.id, app_name=lit_app.name)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass
from typing import Any, Optional

from deepdiff import Delta


@dataclass
class _BaseRequest:
    def to_dict(self):
        return asdict(self)


@dataclass
class _DeltaRequest(_BaseRequest):
    delta: Delta

    def to_dict(self):
        return self.delta.to_dict()


@dataclass
class _CommandRequest(_BaseRequest):
    id: str
    name: str
    method_name: str
    args: Any
    kwargs: Any


@dataclass
class _APIRequest(_BaseRequest):
    id: str
    name: str
    method_name: str
    args: Any
    kwargs: Any


@dataclass
class _RequestResponse(_BaseRequest):
    status_code: int
    content: Optional[str] = None
